import concurrent.futures
from typing import Any

from cachetools.func import ttl_cache
from confluent_kafka.admin import (
    AdminClient,
    ConfigEntry,
    ConfigResource,
    PartitionMetadata,
    TopicMetadata,
)

from streams_explorer.core.config import settings


class KafkaAdminClient:
    def __init__(self) -> None:
        self._enabled: bool = settings.kafka.enable
        self._client: AdminClient
        self.__connect()

    @property
    def enabled(self) -> bool:
        return self._enabled

    def __connect(self) -> None:
        if self.enabled:
            self._client = AdminClient(settings.kafka.config)

    def __describe_config(self, resource: ConfigResource) -> dict:
        return self._client.describe_configs(resources=[resource])

    def __get_resource(self, resource: ConfigResource) -> list[ConfigEntry]:
        fs = self.__describe_config(resource)
        for f in concurrent.futures.as_completed(iter(fs.values())):
            configs = f.result()
            return list(configs.values())
        return []

    def get_topic_config(self, topic: str) -> dict[str, Any]:
        topic_resource = ConfigResource(ConfigResource.Type.TOPIC, topic)
        result = self.__get_resource(topic_resource)
        return self.format_values(result)

    def __get_topic(self, topic: str) -> TopicMetadata | None:
        topics = self._client.list_topics(topic).topics
        return topics.get(topic)

    def get_topic_partitions(self, topic: str) -> dict[int, PartitionMetadata] | None:
        metadata = self.__get_topic(topic)
        if metadata is None:
            return None
        return metadata.partitions

    @ttl_cache(maxsize=1, ttl=settings.kafka.topic_names_cache.ttl)
    def get_all_topic_names(self) -> set[str]:
        return set(self._client.list_topics().topics.keys())

    @staticmethod
    def format_values(values: list[ConfigEntry]) -> dict[str, Any]:
        return {
            config.name: config.value for config in values if not config.is_sensitive
        }



from enum import Enum

from pydantic import BaseModel

from streams_explorer.core.services.dataflow_graph import NodeTypesEnum


class NodeInfoType(str, Enum):
    JSON = "json"
    BASIC = "basic"
    LINK = "link"


class NodeInfoListItem(BaseModel):
    name: str
    value: str | dict
    type: NodeInfoType


class NodeInformation(BaseModel):
    node_id: str
    node_type: NodeTypesEnum
    info: list[NodeInfoListItem]



import re

from cachetools.func import ttl_cache
from fastapi import WebSocket
from kubernetes_asyncio.client import V1beta1CronJob, V1Job
from loguru import logger

from streams_explorer.core.client_manager import ClientManager
from streams_explorer.core.config import settings
from streams_explorer.core.k8s_app import K8sApp
from streams_explorer.core.node_info_extractor import (
    get_displayed_information_connector,
    get_displayed_information_deployment,
    get_displayed_information_topic,
)
from streams_explorer.core.services.dataflow_graph import DataFlowGraph, NodeTypesEnum
from streams_explorer.core.services.kafka_admin_client import KafkaAdminClient
from streams_explorer.core.services.kafkaconnect import KafkaConnect
from streams_explorer.core.services.kubernetes import (
    K8sDeploymentUpdate,
    K8sEvent,
    Kubernetes,
)
from streams_explorer.core.services.linking_services import LinkingService
from streams_explorer.core.services.metric_providers import MetricProvider
from streams_explorer.extractors import extractor_container
from streams_explorer.models.graph import Metric
from streams_explorer.models.k8s import K8sDeploymentUpdateType, K8sReason
from streams_explorer.models.kafka_connector import KafkaConnector
from streams_explorer.models.node_information import (
    NodeInfoListItem,
    NodeInformation,
    NodeInfoType,
)


class StreamsExplorer:
    def __init__(
        self, linking_service: LinkingService, metric_provider: type[MetricProvider]
    ) -> None:
        self.applications: dict[str, K8sApp] = {}
        self.kafka_connectors: list[KafkaConnector] = []
        self.kubernetes = Kubernetes(self)
        self.kafka = KafkaAdminClient()
        self.data_flow = DataFlowGraph(
            metric_provider=metric_provider, kafka=self.kafka
        )
        self.linking_service = linking_service
        self.client_manager = ClientManager()
        self.modified: bool = True

    async def setup(self) -> None:
        await self.kubernetes.setup()

    async def watch(self) -> None:
        await self.kubernetes.watch()

    async def update_graph(self) -> None:
        if not self.modified:
            return  # skip unnecessary re-render
        logger.info("Update graph")
        self.data_flow.reset()
        self.__create_graph()
        self.modified = False
        self.data_flow.setup_metric_provider()
        await self.data_flow.store_json_graph()
        logger.info("Update graph completed")

    def get_positioned_json_graph(self) -> dict:
        return self.data_flow.json_graph

    async def get_positioned_pipeline_json_graph(self, pipeline_name: str) -> dict:
        return await self.data_flow.get_positioned_pipeline_graph(pipeline_name)

    def get_pipeline_names(self) -> list[str]:
        return list(sorted(self.data_flow.pipelines.keys()))

    async def get_metrics(self) -> list[Metric]:
        return await self.data_flow.get_metrics()

    @ttl_cache(ttl=settings.node_info.cache_ttl)
    def get_node_information(self, node_id: str) -> NodeInformation | None:
        node_type = self.data_flow.get_node_type(node_id)

        match node_type:
            case NodeTypesEnum.CONNECTOR:
                config = KafkaConnect.get_connector_config(node_id)
                return NodeInformation(
                    node_id=node_id,
                    node_type=node_type,
                    info=self.linking_service.connector_info
                    + get_displayed_information_connector(config),
                )

            case NodeTypesEnum.TOPIC | NodeTypesEnum.ERROR_TOPIC:
                info = self.linking_service.topic_info
                if self.kafka.enabled:
                    partitions = self.kafka.get_topic_partitions(node_id)
                    if partitions is not None:
                        info.append(
                            NodeInfoListItem(
                                name="Partitions",
                                value=str(len(partitions)),
                                type=NodeInfoType.BASIC,
                            )
                        )
                    config = self.kafka.get_topic_config(node_id)
                    info += get_displayed_information_topic(config)
                info.append(
                    NodeInfoListItem(
                        name="Schema",
                        value={},
                        type=NodeInfoType.JSON,
                    )
                )
                return NodeInformation(
                    node_id=node_id,
                    node_type=node_type,
                    info=info,
                )

            case NodeTypesEnum.STREAMING_APP:
                info = get_displayed_information_deployment(self.applications[node_id])
                return NodeInformation(
                    node_id=node_id,
                    node_type=node_type,
                    info=self.linking_service.streaming_app_info + info,
                )

            case sink_source if sink_source in self.linking_service.sink_source_info:
                return NodeInformation(
                    node_id=node_id,
                    node_type=NodeTypesEnum.SINK_SOURCE,
                    info=self.linking_service.sink_source_info[node_type],
                )

    def get_link(self, node_id: str, link_type: str) -> str | None:
        node_type = self.data_flow.get_node_type(node_id)
        match node_type:
            case NodeTypesEnum.CONNECTOR:
                config = KafkaConnect.get_connector_config(node_id)
                return self.linking_service.get_redirect_connector(config, link_type)
            case NodeTypesEnum.TOPIC | NodeTypesEnum.ERROR_TOPIC:
                return self.linking_service.get_redirect_topic(node_id, link_type)
            case NodeTypesEnum.STREAMING_APP:
                return self.linking_service.get_redirect_streaming_app(
                    self.applications[node_id], link_type
                )
        if node_type in self.linking_service.sink_source_redirects:
            return self.linking_service.get_sink_source_redirects(node_type, node_id)

    async def handle_deployment_update(self, update: K8sDeploymentUpdate) -> None:
        item = update["object"]
        logger.info(
            f"{item.metadata.namespace} {item.__class__.__name__} {update['type']}: {item.metadata.name}"  # pyright: ignore[reportOptionalMemberAccess]
        )

        app: K8sApp | None = None
        match item:
            case V1Job():  # type: ignore[misc]
                app = extractor_container.on_job(item)
            case V1beta1CronJob():  # type: ignore[misc]
                app = extractor_container.on_cron_job(item)
            case _:
                app = K8sApp.factory(item)
        if app:
            await self._handle_app_update(update["type"], app)

    async def _handle_app_update(
        self, type: K8sDeploymentUpdateType, app: K8sApp
    ) -> None:
        match type:
            case K8sDeploymentUpdateType.ADDED | K8sDeploymentUpdateType.MODIFIED:
                await self.__add_app(app)
            case K8sDeploymentUpdateType.DELETED:
                self.__remove_app(app)

    async def handle_event(self, raw_event: K8sEvent) -> None:
        event = raw_event["object"]

        # extract deployment name from pod
        if not event.reason or not event.regarding or not event.regarding.field_path:
            return
        name = re.findall(r"{(.+?)}", event.regarding.field_path)[0]

        logger.info(
            "{} {} {} ({})",
            event.regarding.namespace,
            name,
            event.reason,
            event.type,
        )
        logger.debug(event)

        # map event to application
        if app := self.applications.get(name):
            app.state = K8sReason.from_str(event.reason)
            # app.note = event["note"] # TODO
            await self._update_clients_delta(app)

    def update_connectors(self) -> None:
        extractor_container.reset_connectors()
        logger.info("Retrieve Kafka connectors")
        self.kafka_connectors = KafkaConnect.connectors()
        self.modified = True

    async def update_client_full(self, client: WebSocket) -> None:
        """Send all current application states to client."""
        for app in self.applications.values():
            await self.client_manager.send(client, app.to_state_update())

    async def _update_clients_delta(self, app: K8sApp) -> None:
        """Broadcast a new application state to clients."""
        await self.client_manager.broadcast(app.to_state_update())

    async def __add_app(self, app: K8sApp) -> None:
        if app.is_streams_app():
            self.applications[app.id] = app
            self.modified = True
            extractor_container.on_streaming_app_add(app.config)
            await self._update_clients_delta(app)

    def __remove_app(self, app: K8sApp) -> None:
        if app.is_streams_app():
            self.applications.pop(app.id)
            self.modified = True
            extractor_container.on_streaming_app_delete(app.config)

    def __create_graph(self) -> None:
        logger.info("Setup pipeline graph")
        for _, app in self.applications.items():
            self.data_flow.add_streaming_app(app)

        for connector in self.kafka_connectors:
            self.data_flow.add_connector(connector)

        sources, sinks = extractor_container.get_sources_sinks()
        for source in sources:
            self.data_flow.add_source(source)

        for sink in sinks:
            self.data_flow.add_sink(sink)

        self.data_flow.apply_input_pattern_edges()



from fastapi import APIRouter, Depends, HTTPException
from loguru import logger
from starlette import status

from streams_explorer.api.dependencies.streams_explorer import get_streams_explorer
from streams_explorer.core.services.schemaregistry import SchemaRegistry
from streams_explorer.models.node_information import NodeInformation
from streams_explorer.streams_explorer import StreamsExplorer

router = APIRouter()


@router.get("/{node_id}", response_model=NodeInformation)
async def get_node_info(
    node_id: str,
    streams_explorer: StreamsExplorer = Depends(get_streams_explorer),
):
    try:
        return streams_explorer.get_node_information(node_id)
    except Exception:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f'Could not find information for node with name "{node_id}"',
        )


@router.get("/{node_id}/schema", response_model=list[int])
async def get_node_schema_versions(node_id: str):
    return SchemaRegistry.get_versions(node_id)


@router.get("/{node_id}/schema/{version}", response_model=dict)
async def get_node_schema(node_id: str, version: int):
    return SchemaRegistry.get_schema(node_id, version=version)


@router.get("/linking/{node_id}", response_model=str)
def get_linking(
    node_id: str,
    link_type: str,
    streams_explorer: StreamsExplorer = Depends(get_streams_explorer),
):
    url = streams_explorer.get_link(node_id, link_type)
    logger.info(f"Redirecting to {url}")
    return url



from __future__ import annotations

from typing import TYPE_CHECKING

from kubernetes_asyncio.client import V1beta1CronJob, V1Job

import streams_explorer.core.k8s_app as k8s
from streams_explorer.core.extractor.extractor import ProducerAppExtractor

if TYPE_CHECKING:
    from streams_explorer.core.k8s_app import K8sAppCronJob, K8sAppJob


class StreamsBootstrapProducer(ProducerAppExtractor):
    def on_job_parsing(self, job: V1Job) -> K8sAppJob | None:
        producer = k8s.K8sAppJob(job)

        # filter out Jobs created by CronJobs
        if producer.metadata.owner_references:
            for owner_reference in producer.metadata.owner_references:
                if owner_reference.kind == "CronJob":
                    return None

        if producer.is_streams_app():
            return producer

    def on_cron_job_parsing(self, cron_job: V1beta1CronJob) -> K8sAppCronJob | None:
        producer = k8s.K8sAppCronJob(cron_job)
        if producer.is_streams_app():
            return producer



from streams_explorer.core.config import settings
from streams_explorer.core.k8s_app import K8sApp
from streams_explorer.core.services.linking_services import LinkingService
from streams_explorer.models.node_information import NodeInfoListItem, NodeInfoType


class DefaultLinker(LinkingService):
    def __init__(self) -> None:
        super().__init__()
        self.sink_source_redirects = {"elasticsearch-index"}
        self.sink_source_info = {
            "elasticsearch-index": [
                NodeInfoListItem(name="Kibana", value="", type=NodeInfoType.LINK)
            ]
        }

        if settings.grafana.enable:
            self.add_topic_info_item(
                NodeInfoListItem(
                    name="Topic Monitoring", value="grafana", type=NodeInfoType.LINK
                )
            )
            grafana_consumer_link = NodeInfoListItem(
                name="Consumer Group Monitoring",
                value="grafana",
                type=NodeInfoType.LINK,
            )
            self.add_streaming_app_info_item(grafana_consumer_link)
            self.add_connector_info_item(grafana_consumer_link)

        if settings.akhq.enable:
            self.add_message_provider("akhq")
            if settings.akhq.get("connect"):
                self.add_connector_info_item(
                    NodeInfoListItem(
                        name="Connector Tasks",
                        value="akhq-connect",
                        type=NodeInfoType.LINK,
                    )
                )

        if settings.redpanda_console.enable:
            self.add_message_provider("redpanda_console")

        if settings.kibanalogs.enable:
            self.add_logging_provider("kibanalogs")

        if settings.loki.enable:
            self.add_logging_provider("loki")

    def get_redirect_connector(self, config: dict, link_type: str) -> str | None:
        if connector_name := config.get("name"):
            consumer_group = f"connect-{connector_name}"
            match link_type:
                case "grafana":
                    return f"{settings.grafana.url}/d/{settings.grafana.dashboards.consumergroups}?var-consumergroups={consumer_group}"
                case "akhq":
                    return f"{settings.akhq.url}/ui/{settings.akhq.cluster}/group/{consumer_group}"
                case "akhq-connect":
                    return f"{settings.akhq.url}/ui/{settings.akhq.cluster}/connect/{settings.akhq.get('connect')}/definition/{connector_name}/tasks"
                case "redpanda_console":
                    return f"{settings.redpanda_console.url}/groups/{consumer_group}"

    def get_redirect_topic(self, topic_name: str, link_type: str) -> str | None:
        match link_type:
            case "grafana":
                return f"{settings.grafana.url}/d/{settings.grafana.dashboards.topics}?var-topics={topic_name}"
            case "akhq":
                return (
                    f"{settings.akhq.url}/ui/{settings.akhq.cluster}/topic/{topic_name}"
                )
            case "redpanda_console":
                return f"{settings.redpanda_console.url}/topics/{topic_name}"

    def get_redirect_streaming_app(self, k8s_app: K8sApp, link_type: str) -> str | None:
        consumer_group = k8s_app.consumer_group
        match link_type:
            case "kibanalogs":
                return f"{settings.kibanalogs.url}/app/discover#/?_a=(columns:!(message),query:(language:lucene,query:'kubernetes.labels.app: \"{k8s_app.name}\"'))"
            case "loki":
                return f'{settings.loki.url}/explore?orgId=1&left=["now-1h","now","loki",{{"expr":"{{app=\\"{k8s_app.name}\\"}}"}}]'
            case "grafana" if consumer_group:
                return f"{settings.grafana.url}/d/{settings.grafana.dashboards.consumergroups}?var-consumergroups={consumer_group}"
            case "akhq" if consumer_group:
                return f"{settings.akhq.url}/ui/{settings.akhq.cluster}/group/{consumer_group}"
            case "redpanda_console" if consumer_group:
                return f"{settings.redpanda_console.url}/groups/{consumer_group}"

    def get_sink_source_redirects(
        self, node_type: str, _sink_source_name: str
    ) -> str | None:
        if node_type == "elasticsearch-index":
            return settings.esindex.url

    def add_message_provider(self, value: str) -> None:
        self.add_topic_info_item(
            NodeInfoListItem(name="Message Viewer", value=value, type=NodeInfoType.LINK)
        )
        consumer_link = NodeInfoListItem(
            name="Consumer Group Details", value=value, type=NodeInfoType.LINK
        )
        self.add_streaming_app_info_item(consumer_link)
        self.add_connector_info_item(consumer_link)

    def add_logging_provider(self, value: str) -> None:
        self.add_streaming_app_info_item(
            NodeInfoListItem(name="Logs", value=value, type=NodeInfoType.LINK),
        )



from fastapi import FastAPI
from starlette.requests import Request
from starlette.websockets import WebSocket

from streams_explorer.streams_explorer import StreamsExplorer

# FIXME: Pydantic bug
# @overload
# def get_application(*, request: Request, websocket=None) -> FastAPI:
#     ...


# @overload
# def get_application(*, request=None, websocket: WebSocket) -> FastAPI:
#     ...


def get_application(
    *, request: Request | None = None, websocket: WebSocket | None = None
) -> FastAPI:
    if request:
        return request.app
    if websocket:
        return websocket.app
    raise ValueError("no request or websocket argument given")


# Pydantic doesn't support correct type annotations (Request | None & WebSocket | None)
def get_streams_explorer(*, request: Request = None, websocket: WebSocket = None) -> StreamsExplorer:  # type: ignore
    app = get_application(request=request, websocket=websocket)
    return get_streams_explorer_from_state(app)


def get_streams_explorer_from_state(app: FastAPI) -> StreamsExplorer:
    return app.state.streams_explorer



from loguru import logger

from streams_explorer.core.services.linking_services import LinkingService
from streams_explorer.defaultlinker import DefaultLinker
from streams_explorer.plugins import load_plugin


def load_linker() -> type[LinkingService]:
    linker = load_plugin(LinkingService)  # type: ignore[misc]
    if not linker:
        logger.info("Using default LinkingService")
        return DefaultLinker
    logger.info(f"Using custom LinkingService: {linker.__name__}")
    return linker



from fastapi import APIRouter

from streams_explorer.api.routes import (
    graph,
    kubernetes_health,
    metrics,
    node,
    pipelines,
    update,
)

router = APIRouter()

router.include_router(update.router, prefix="/update")
router.include_router(graph.router, prefix="/graph")
router.include_router(pipelines.router, prefix="/pipelines")
router.include_router(node.router, prefix="/node")
router.include_router(metrics.router, prefix="/metrics")
router.include_router(kubernetes_health.router, prefix="/health")



from starlette.requests import Request

from streams_explorer.core.services.dataflow_graph import DataFlowGraph


def get_dataflow_graph(request: Request) -> DataFlowGraph:
    return request.app.state.dataflow_graph



from dataclasses import dataclass, field

from fastapi import WebSocket
from loguru import logger
from pydantic import BaseModel


@dataclass(frozen=True)
class ClientManager:
    _clients: list[WebSocket] = field(init=False, default_factory=list)

    async def connect(self, websocket: WebSocket) -> None:
        await websocket.accept()
        logger.info("WebSocket client {} connected", websocket.client)
        self._clients.append(websocket)

    async def disconnect(self, websocket: WebSocket) -> None:
        await websocket.close()
        logger.info("WebSocket client {} disconnected", websocket.client)
        self._clients.remove(websocket)

    async def send(self, websocket: WebSocket, obj: BaseModel) -> None:
        await websocket.send_json(obj.dict())

    async def broadcast(self, obj: BaseModel) -> None:
        data = obj.dict()
        for client in self._clients:
            await client.send_json(data)

    @property
    def clients(self) -> list[WebSocket]:
        return self._clients



from fastapi import APIRouter, Depends

from streams_explorer.api.dependencies.streams_explorer import get_streams_explorer
from streams_explorer.models.graph import Metric
from streams_explorer.streams_explorer import StreamsExplorer

router = APIRouter()


@router.get("", response_model=list[Metric])
async def get_metrics(
    streams_explorer: StreamsExplorer = Depends(get_streams_explorer),
):
    return await streams_explorer.get_metrics()



from fastapi import APIRouter, Depends

from streams_explorer.api.dependencies.streams_explorer import get_streams_explorer
from streams_explorer.models.pipelines import Pipelines
from streams_explorer.streams_explorer import StreamsExplorer

router = APIRouter()


@router.get("", response_model=Pipelines)
def get_pipelines(
    streams_explorer: StreamsExplorer = Depends(get_streams_explorer),
):
    return Pipelines(pipelines=streams_explorer.get_pipeline_names())



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Awaitable, Callable, NamedTuple, TypedDict

import kubernetes_asyncio.client
import kubernetes_asyncio.config
import kubernetes_asyncio.watch
from kubernetes_asyncio.client import (
    ApiException,
    EventsV1Event,
    EventsV1EventList,
    V1beta1CronJob,
    V1beta1CronJobList,
    V1Deployment,
    V1DeploymentList,
    V1Job,
    V1JobList,
    V1StatefulSet,
    V1StatefulSetList,
)
from loguru import logger

from streams_explorer.core.config import settings
from streams_explorer.core.k8s_app import K8sObject
from streams_explorer.models.k8s import K8sDeploymentUpdateType, K8sEventType

if TYPE_CHECKING:
    from streams_explorer.streams_explorer import StreamsExplorer


class K8sResource(NamedTuple):
    func: Callable[
        ...,
        V1DeploymentList
        | V1StatefulSetList
        | V1JobList
        | V1beta1CronJobList
        | EventsV1EventList,
    ]
    return_type: type | None
    callback: Callable[..., Awaitable[None]]
    delay: int = 0


class K8sDeploymentUpdate(TypedDict):
    type: K8sDeploymentUpdateType
    object: K8sObject


class K8sEvent(TypedDict):
    type: K8sEventType
    object: EventsV1Event


class Kubernetes:
    context: str = settings.k8s.deployment.context
    namespaces: list[str] = settings.k8s.deployment.namespaces

    def __init__(self, streams_explorer: StreamsExplorer) -> None:
        self.streams_explorer = streams_explorer

    async def setup(self) -> None:
        try:
            if settings.k8s.deployment.cluster:
                logger.info("Setup K8s environment in cluster")
                kubernetes_asyncio.config.load_incluster_config()
            else:
                logger.info("Setup K8s environment")
                await kubernetes_asyncio.config.load_kube_config(context=self.context)
        except kubernetes_asyncio.config.ConfigException as e:
            raise Exception("Could not load K8s environment configuration") from e

        conf = kubernetes_asyncio.client.Configuration.get_default_copy()
        conf.client_side_validation = False
        kubernetes_asyncio.client.Configuration.set_default(conf)

        self.k8s_app_client = kubernetes_asyncio.client.AppsV1Api()
        self.k8s_batch_client = kubernetes_asyncio.client.BatchV1Api()
        self.k8s_beta_batch_client = kubernetes_asyncio.client.BatchV1beta1Api()
        self.k8s_events_client = kubernetes_asyncio.client.EventsV1Api()

    async def watch(self) -> None:
        def list_deployments(namespace: str, *args, **kwargs) -> V1DeploymentList:
            return self.k8s_app_client.list_namespaced_deployment(
                *args, namespace=namespace, **kwargs
            )

        def list_stateful_sets(namespace: str, *args, **kwargs) -> V1StatefulSetList:
            return self.k8s_app_client.list_namespaced_stateful_set(
                *args, namespace=namespace, **kwargs
            )

        def list_jobs(namespace: str, *args, **kwargs) -> V1JobList:
            return self.k8s_batch_client.list_namespaced_job(
                *args, namespace=namespace, **kwargs
            )

        def list_cron_jobs(namespace: str, *args, **kwargs) -> V1beta1CronJobList:
            return self.k8s_beta_batch_client.list_namespaced_cron_job(
                *args, namespace=namespace, **kwargs
            )

        def list_events(namespace: str, *args, **kwargs) -> EventsV1EventList:
            return self.k8s_events_client.list_namespaced_event(
                *args, namespace=namespace, **kwargs
            )

        resources = (
            K8sResource(
                list_deployments,
                V1Deployment,
                self.streams_explorer.handle_deployment_update,
            ),
            K8sResource(
                list_stateful_sets,
                V1StatefulSet,
                self.streams_explorer.handle_deployment_update,
            ),
            K8sResource(
                list_jobs,
                V1Job,
                self.streams_explorer.handle_deployment_update,
            ),
            K8sResource(
                list_cron_jobs,
                V1beta1CronJob,
                self.streams_explorer.handle_deployment_update,
            ),
            K8sResource(
                list_events,
                EventsV1Event,
                self.streams_explorer.handle_event,
                delay=5,
            ),
        )

        for resource in resources:
            await asyncio.sleep(resource.delay)
            for namespace in self.namespaces:
                asyncio.create_task(
                    self.__watch_namespace(
                        namespace,
                        resource,
                    )
                )

    async def __watch_namespace(
        self,
        namespace: str,
        resource: K8sResource,
    ) -> None:
        return_type = resource.return_type.__name__ if resource.return_type else None
        try:
            async with kubernetes_asyncio.watch.Watch(return_type) as w:
                async with w.stream(resource.func, namespace) as stream:
                    async for event in stream:
                        await resource.callback(event)
        except ApiException as e:
            if e.status == 410:
                # restart watch to get fresh resource version
                return await self.__watch_namespace(namespace, resource)
            else:
                raise



from __future__ import annotations

from abc import abstractmethod
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from kubernetes_asyncio.client import V1beta1CronJob, V1Job

from streams_explorer.models.k8s import K8sConfig
from streams_explorer.models.kafka_connector import KafkaConnector
from streams_explorer.models.sink import Sink
from streams_explorer.models.source import Source
from streams_explorer.plugins import Plugin

if TYPE_CHECKING:
    from streams_explorer.core.k8s_app import K8sAppCronJob, K8sAppJob


@dataclass
class Extractor(Plugin):
    sources: list[Source] = field(default_factory=list)
    sinks: list[Sink] = field(default_factory=list)

    def reset(self) -> None:
        self.sources.clear()
        self.sinks.clear()


class ConnectorExtractor(Extractor):
    @abstractmethod
    def on_connector_info_parsing(
        self, info: dict, connector_name: str
    ) -> KafkaConnector | None:
        ...


class StreamsAppExtractor(Extractor):
    @abstractmethod
    def on_streaming_app_add(self, config: K8sConfig) -> None:
        ...

    @abstractmethod
    def on_streaming_app_delete(self, config: K8sConfig) -> None:
        ...


class ProducerAppExtractor(Extractor):
    @abstractmethod
    def on_job_parsing(self, job: V1Job) -> K8sAppJob | None:
        ...

    @abstractmethod
    def on_cron_job_parsing(self, cron_job: V1beta1CronJob) -> K8sAppCronJob | None:
        ...



from typing import TypeVar

from loguru import logger

from streams_explorer.core.config import settings
from streams_explorer.core.k8s_app import K8sApp
from streams_explorer.models.node_information import NodeInfoListItem, NodeInfoType

V = TypeVar("V")


def find(element: str, json: dict) -> dict:
    keys = element.split(".")
    key_value = json
    for key in keys:
        key_value = key_value[key]
    return key_value


def get_info(key: str, config: dict[str, V]) -> V | dict[str, V] | None:
    if value := config.get(key):
        return value

    try:
        return find(key, config)
    except KeyError:
        return None


def get_type(value) -> NodeInfoType:
    if isinstance(value, dict):
        return NodeInfoType.JSON
    return NodeInfoType.BASIC


def get_displayed_info(
    displayed_info_settings: list[dict[str, str]], config: dict
) -> list[NodeInfoListItem]:
    node_infos = []
    for item in displayed_info_settings:
        name = item.get("name")
        key = item.get("key")
        if not name or not key:
            continue

        value = get_info(key, config)
        if value is None:
            logger.warning(f'Could not find key "{key}"')
        else:
            value_type = get_type(value)
            if value_type != NodeInfoType.JSON:
                value = str(value)
            node_infos.append(NodeInfoListItem(name=name, value=value, type=value_type))
    return node_infos


def get_displayed_information_connector(config: dict) -> list[NodeInfoListItem]:
    connector_settings = settings.kafkaconnect.displayed_information
    return get_displayed_info(connector_settings, config)


def get_displayed_information_deployment(k8s: K8sApp) -> list[NodeInfoListItem]:
    k8s_settings = settings.k8s.displayed_information
    return get_displayed_info(k8s_settings, k8s.to_dict())


def get_displayed_information_topic(config: dict) -> list[NodeInfoListItem]:
    kafka_settings = settings.kafka.displayed_information
    return get_displayed_info(kafka_settings, config)



from __future__ import annotations

import json
from functools import wraps
from typing import Callable, ParamSpec, TypeVar

import httpx
from loguru import logger

from streams_explorer.core.config import settings
from streams_explorer.core.services.dataflow_graph import NodeNotFound

url: str | None = settings.schemaregistry.url

T = TypeVar("T")
P = ParamSpec("P")


def default_return(func: Callable[P, T]) -> Callable[P, T]:
    """Decorator which returns an empty instance of the function's return type, if Schema Registry is disabled."""

    @wraps(func)
    def inner(*args: P.args, **kwargs: P.kwargs) -> T:
        if url is None:
            try:
                typ = eval(func.__annotations__["return"])
                return typ()
            except KeyError:
                raise Exception(f"'{func.__name__}' is missing return type annotation")
        return func(*args, **kwargs)

    return inner


class SchemaRegistry:
    @staticmethod
    @default_return
    def get_versions(topic: str) -> list[int]:
        logger.info(f"Fetch schema versions for topic {topic}")
        response = httpx.get(f"{url}/subjects/{topic}-value/versions/")
        if response.status_code == 200:
            data = response.json()
            return data
        logger.debug(f"Error fetching schema versions for topic {topic}: {response}")
        return []

    @staticmethod
    @default_return
    def get_schema(topic: str, version: int = 1) -> dict:
        try:
            logger.info(f"Fetch schema version {version} for {topic}")
            response = httpx.get(f"{url}/subjects/{topic}-value/versions/{version}")
            return json.loads(response.json().get("schema"))
        except Exception:
            raise NodeNotFound()



from typing import Any, TypeAlias

from dynaconf import Dynaconf, Validator

APP_NAME = "Streams Explorer"
API_PREFIX = "/api"

settings: Any = Dynaconf(
    envvar_prefix="se",
    settings_files=["settings.yaml"],
    load_dotenv=True,
    validators=[
        Validator("graph.update_interval", must_exist=True, is_type_of=int),
        Validator("graph.layout_arguments", must_exist=True, is_type_of=str),
        Validator("graph.pipeline_distance", must_exist=True, is_type_of=int),
        Validator(
            "graph.resolve.input_pattern_topics.all", must_exist=True, is_type_of=bool
        ),
        Validator(
            "graph.resolve.input_pattern_topics.pipelines",
            must_exist=True,
            is_type_of=bool,
        ),
        Validator("k8s.deployment.cluster", must_exist=True, is_type_of=bool),
        Validator("k8s.deployment.context", is_type_of=str),
        Validator("k8s.deployment.namespaces", must_exist=True, is_type_of=list),
        Validator("k8s.containers.ignore", is_type_of=list),
        Validator("k8s.labels", must_exist=True, is_type_of=list),
        Validator("k8s.pipeline.label", must_exist=True, is_type_of=str),
        Validator("k8s.consumer_group_annotation", must_exist=True, is_type_of=str),
        Validator("kafka.enable", eq=False, default=False)
        | (
            Validator("kafka.enable", eq=True)
            & Validator(
                "kafka.config",
                is_type_of=dict,
                must_exist=True,
                condition=lambda v: type(v.get("bootstrap.servers")) is str,
            )
        ),
        Validator("kafka.displayed_information", is_type_of=list, default=[]),
        Validator("kafka.topic_names_cache.ttl", is_type_of=int, default=3600),
        Validator("node_info.cache_ttl", must_exist=True, is_type_of=int),
        Validator("kafkaconnect.url", default=None),
        Validator("kafkaconnect.update_interval", is_type_of=int, default=300),
        Validator("kafkaconnect.displayed_information", is_type_of=list, default=[]),
        Validator("schemaregistry.url", default=None),
        Validator("prometheus.url", must_exist=True, is_type_of=str),
        Validator("grafana.enable", is_type_of=bool, default=False),
        Validator("grafana.dashboards.topics", is_type_of=str),
        Validator("grafana.dashboards.consumergroups", is_type_of=str),
        Validator("akhq.enable", is_type_of=bool, default=False),
        Validator("redpanda_console.enable", is_type_of=bool, default=False),
        Validator("akhq.enable", eq=False)
        | Validator("redpanda_console.enable", eq=False),
        Validator("kibanalogs.enable", is_type_of=bool, default=False),
        Validator("loki.enable", is_type_of=bool, default=False),
        Validator("kibanalogs.enable", eq=False) | Validator("loki.enable", eq=False),
        Validator("plugins.path", must_exist=True, is_type_of=str),
        Validator("plugins.extractors.default", must_exist=True, is_type_of=bool),
    ],
)

InfoList: TypeAlias = list[dict[str, str]]


def sort_displayed_information(lst: InfoList) -> InfoList:
    return sorted(lst, key=lambda k: str.casefold(k["name"]))


settings.k8s.displayed_information = sort_displayed_information(
    settings.k8s.displayed_information
)
settings.kafka.displayed_information = sort_displayed_information(
    settings.kafka.displayed_information
)
settings.kafkaconnect.displayed_information = sort_displayed_information(
    settings.kafkaconnect.displayed_information
)



from __future__ import annotations

import importlib
import inspect
import sys
from abc import ABC
from collections.abc import Sequence
from pathlib import Path
from types import ModuleType
from typing import Literal, TypeVar, overload

from loguru import logger

from streams_explorer.core.config import settings


class Plugin(ABC):
    """Plugin base class."""

    ...


T = TypeVar("T", bound=Plugin)  # Plugin type


@overload
def load_plugin(base_class: type[T], all: Literal[False] = False) -> type[T] | None:
    ...


@overload
def load_plugin(base_class: type[T], all: Literal[True]) -> Sequence[type[T]]:
    ...


def load_plugin(
    base_class: type[T], all: bool = False
) -> type[T] | Sequence[type[T]] | None:
    if not settings.plugins.path:
        return None
    path = Path(settings.plugins.path)
    sys.path.append(str(path))
    logger.info(f"Loading {base_class} from {path}")
    modules: list[type[T]] = []
    for file in path.glob("*.py"):
        module = importlib.import_module(file.stem)
        plugin_class = _find_class(module, base_class)
        if plugin_class is None:
            continue
        logger.info(f"Found {plugin_class} {file}")
        if not all:
            return plugin_class
        modules.append(plugin_class)
    return modules


def _find_class(module: ModuleType, base_class: type[T]) -> type[T] | None:
    members = inspect.getmembers(module, inspect.isclass)
    if not members:
        return None
    if len(members) > 1:  # multiple classes found
        # exclude classes imported from other modules
        members = [
            (name, val) for name, val in members if val.__module__ == module.__name__
        ]
    name, _ = members[0]
    plugin_class = getattr(module, name)
    if (
        plugin_class
        and issubclass(plugin_class, base_class)
        and plugin_class is not base_class
    ):
        return plugin_class



from enum import Enum


class NodeTypesEnum(str, Enum):
    STREAMING_APP = "streaming-app"
    CONNECTOR = "connector"
    TOPIC = "topic"
    ERROR_TOPIC = "error-topic"
    SINK_SOURCE = "sink/source"



import re
from collections import defaultdict
from enum import Enum

import networkx as nx
from loguru import logger
from networkx.drawing.nx_agraph import graphviz_layout
from networkx.generators.ego import ego_graph

from streams_explorer.core.config import settings
from streams_explorer.core.k8s_app import ATTR_PIPELINE, K8sApp
from streams_explorer.core.services.kafka_admin_client import KafkaAdminClient
from streams_explorer.core.services.metric_providers import MetricProvider
from streams_explorer.models.graph import GraphEdge, GraphNode, Metric
from streams_explorer.models.kafka_connector import (
    KafkaConnector,
    KafkaConnectorTypesEnum,
)
from streams_explorer.models.node_types import NodeTypesEnum
from streams_explorer.models.sink import Sink
from streams_explorer.models.source import Source


class NodeNotFound(Exception):
    pass


class PipelineNotFound(Exception):
    pass


class NodeDataFields(str, Enum):
    NODE_TYPE = "node_type"
    LABEL = "label"


class DataFlowGraph:
    def __init__(
        self, metric_provider: type[MetricProvider], kafka: KafkaAdminClient
    ) -> None:
        self.graph = nx.DiGraph()
        self.json_graph: dict = {}
        self.pipelines: dict[str, nx.DiGraph] = {}
        self.json_pipelines: dict[str, dict] = {}
        self.metric_provider_class = metric_provider
        self.metric_provider: MetricProvider | None = None
        self.kafka = kafka

        self._topic_pattern_queue: defaultdict[str, set[str]] = defaultdict(
            set
        )  # topic pattern -> set of target node ids

    async def store_json_graph(self) -> None:
        self.json_graph = await self.get_positioned_graph()

    def setup_metric_provider(self) -> None:
        self.metric_provider = self.metric_provider_class(
            list(self.graph.nodes(data=True))
        )

    def add_streaming_app(self, app: K8sApp) -> None:
        pipeline = app.attributes.get(ATTR_PIPELINE)

        self._add_streaming_app(self.graph, app)
        if pipeline:
            self._add_streaming_app(
                self.pipelines.setdefault(pipeline, nx.DiGraph()), app
            )

    def _add_streaming_app(self, graph: nx.DiGraph, app: K8sApp) -> None:
        graph.add_node(
            app.id,
            label=app.name,
            node_type=NodeTypesEnum.STREAMING_APP,
            **app.attributes,
        )

        for input_topic in app.input_topics:
            self._add_topic(graph, input_topic)
            self._add_input_topic(graph, app.id, input_topic)
        if app.output_topic:
            self._add_topic(graph, app.output_topic)
            self._add_output_topic(graph, app.id, app.output_topic)
        if app.error_topic:
            self._add_error_topic(graph, app.id, app.error_topic)
        if app.input_pattern:
            self._enqueue_input_pattern(app.input_pattern, app.id)
        for extra_input in app.extra_input_topics:
            self._add_topic(graph, extra_input)
            self._add_input_topic(graph, app.id, extra_input)
        for extra_output in app.extra_output_topics:
            self._add_topic(graph, extra_output)
            self._add_output_topic(graph, app.id, extra_output)
        for extra_pattern in app.extra_input_patterns:
            self._enqueue_input_pattern(extra_pattern, app.id)

    def add_connector(
        self, connector: KafkaConnector, pipeline: str | None = None
    ) -> None:
        graph = self.graph
        if pipeline is not None:
            graph = self.pipelines[pipeline]

        graph.add_node(
            connector.name,
            label=connector.name,
            node_type=NodeTypesEnum.CONNECTOR,
        )
        for topic in connector.get_topics():
            self._add_topic(graph, topic)
            if connector.type == KafkaConnectorTypesEnum.SINK:
                graph.add_edge(topic, connector.name)
            elif connector.type == KafkaConnectorTypesEnum.SOURCE:
                graph.add_edge(connector.name, topic)
        if connector_error_topic := connector.get_error_topic():
            self._add_error_topic(graph, connector.name, connector_error_topic)

        # Add to pipeline graph
        if pipeline is None:
            if pipelines := self.find_associated_pipelines(
                connector.name,
                reverse=connector.type == KafkaConnectorTypesEnum.SINK,
                radius=2,
            ):
                for pipeline in pipelines:
                    self.add_connector(connector, pipeline=pipeline)

    def add_source(self, source: Source) -> None:
        node: GraphNode = (
            source.name,
            {
                NodeDataFields.LABEL: source.name,
                NodeDataFields.NODE_TYPE: source.node_type,
            },
        )
        edge: GraphEdge = (source.name, source.target)
        self.add_to_graph(node, edge)

    def add_sink(self, sink: Sink) -> None:
        node: GraphNode = (
            sink.name,
            {NodeDataFields.LABEL: sink.name, NodeDataFields.NODE_TYPE: sink.node_type},
        )
        edge: GraphEdge = (sink.source, sink.name)
        self.add_to_graph(node, edge, reverse=True)

    def add_to_graph(
        self, node: GraphNode, edge: GraphEdge, reverse: bool = False
    ) -> None:
        node_name, node_data = node
        self.graph.update(nodes=[node], edges=[edge])

        if pipelines := self.find_associated_pipelines(node_name, reverse=reverse):
            target = (set(edge) - {node_name}).pop()
            for pipeline in pipelines:
                # verify target exists in pipeline graph
                if not self.pipelines[pipeline].has_node(target):
                    logger.debug(
                        f"'{node_name}' doesn't belong to pipeline '{pipeline}', '{target}' is not a member of graph"
                    )
                    continue
                self.pipelines[pipeline].add_node(node_name, **node_data)
                self.pipelines[pipeline].add_edge(*edge)

    async def get_positioned_pipeline_graph(self, pipeline_name: str) -> dict:
        if pipeline_name not in self.pipelines:
            raise PipelineNotFound()
        # caching
        if pipeline_name not in self.json_pipelines:
            self.json_pipelines[pipeline_name] = await self.__get_positioned_json_graph(
                self.pipelines[pipeline_name]
            )
        return self.json_pipelines[pipeline_name]

    async def get_positioned_graph(self) -> dict:
        return await self.__get_positioned_json_graph(self.graph)

    async def get_metrics(self) -> list[Metric]:
        if self.metric_provider is not None:
            return await self.metric_provider.get()
        return []

    def get_node_type(self, id: str) -> NodeTypesEnum:
        try:
            return self.graph.nodes[id][NodeDataFields.NODE_TYPE]
        except KeyError:
            raise NodeNotFound()

    def find_associated_pipelines(
        self, node_name: str, reverse: bool = False, radius: int = 3
    ) -> set[str]:
        """
        Search neighborhood of connected successor nodes for pipeline label (used for sources).
        With reverse=True the neighborhood of predecessor nodes is searched instead (used for sinks).
        """
        graph = self.graph.reverse() if reverse else self.graph
        neighborhood = ego_graph(graph, node_name, radius=radius, undirected=False)
        pipelines = set()
        for _, node in neighborhood.nodes(data=True):
            if pipeline := node.get(ATTR_PIPELINE):
                pipelines.add(pipeline)
        if pipelines:
            logger.debug("Pipelines found for {}: {}", node_name, pipelines)
        else:
            logger.warning("No pipeline found for {}", node_name)
        return pipelines

    @staticmethod
    def _add_topic(graph: nx.DiGraph, name: str) -> None:
        graph.add_node(name, label=name, node_type=NodeTypesEnum.TOPIC)

    @staticmethod
    def _filter_topic_node_ids(graph: nx.DiGraph) -> set[str]:
        return {
            node_id
            for node_id, data in graph.nodes(data=True)
            if data[  # pyright: ignore[reportOptionalSubscript]
                NodeDataFields.NODE_TYPE
            ]
            in (NodeTypesEnum.TOPIC, NodeTypesEnum.ERROR_TOPIC)
        }

    @staticmethod
    def _add_input_topic(graph: nx.DiGraph, app_id: str, topic_name: str) -> None:
        graph.add_edge(topic_name, app_id)

    def _add_output_topic(
        self,
        graph: nx.DiGraph,
        app_id: str,
        topic_name: str,
    ) -> None:
        self._add_topic(graph, topic_name)
        graph.add_edge(app_id, topic_name)

    def _enqueue_input_pattern(self, pattern: str, node_id: str) -> None:
        """
        Enqueue a input topic pattern for an app or Kafka Connector
        """
        self._topic_pattern_queue[pattern].add(node_id)

    def apply_input_pattern_edges(self) -> None:
        topics = DataFlowGraph._filter_topic_node_ids(self.graph)
        kafka_topics = self.kafka.get_all_topic_names() if self.kafka.enabled else set()
        for pattern, node_ids in self._topic_pattern_queue.items():
            regex = re.compile(pattern)
            matching_graph_known_topics = set(filter(regex.match, topics))
            # for unknown topics (unkown means not already present in the graph)
            # in the graph we have to create the topic node
            matching_unknown_kafka_topics = set(
                filter(regex.match, kafka_topics)
            ).difference(matching_graph_known_topics)
            for node_id in node_ids:  # node_id can be an app or a kafka connector
                pipeline = self.graph.nodes[node_id].get(ATTR_PIPELINE)
                # handle matching topics that are already present in the graph
                self.handle_matching_topics(
                    matching_graph_known_topics,
                    node_id,
                    pattern,
                    pipeline,
                    add_topic=False,
                )
                # handle unknown topics that are not present in the graph
                self.handle_matching_topics(
                    matching_unknown_kafka_topics,
                    node_id,
                    pattern,
                    pipeline,
                    add_topic=True,
                )

    def handle_matching_topics(
        self,
        matching_unknown_kafka_topics: set[str],
        node_id: str,
        pattern: str,
        pipeline: str | None,
        add_topic: bool = False,
    ) -> None:
        for matched_topic in matching_unknown_kafka_topics:
            if add_topic:
                self._add_topic(self.graph, matched_topic)
            if pipeline is not None:
                self.resolve_topic_pattern_in_pipeline(
                    matched_topic, node_id, pipeline, pattern
                )
            self.resolve_topic_pattern_in_all_graph(matched_topic, node_id, pattern)

    def resolve_topic_pattern_in_all_graph(
        self, matched_topic: str, node_id: str, pattern: str
    ) -> None:
        # resolve topic pattern in overall graph containing all pipelines
        if settings.graph.resolve.input_pattern_topics.all:
            # connect topic to graph
            self.graph.add_edge(matched_topic, node_id)
        else:
            self.add_pattern_as_topic(self.graph, node_id, pattern)

    def add_pattern_as_topic(
        self, graph: nx.DiGraph, node_id: str, pattern: str
    ) -> None:
        # visualize the pattern as topic
        self._add_topic(graph, pattern)
        graph.add_edge(pattern, node_id)

    def resolve_topic_pattern_in_pipeline(
        self, matched_topic: str, node_id: str, pipeline: str, pattern: str
    ) -> None:
        # resolve topic patterns in pipelines
        if settings.graph.resolve.input_pattern_topics.pipelines:
            if matched_topic not in self.pipelines[pipeline].nodes:
                node = self.graph.nodes[matched_topic]
                self.pipelines[pipeline].add_node(matched_topic, **node)
            self.pipelines[pipeline].add_edge(matched_topic, node_id)
        else:
            self.add_pattern_as_topic(self.pipelines[pipeline], node_id, pattern)

    @staticmethod
    def _add_error_topic(
        graph: nx.DiGraph,
        app_id: str,
        topic_name: str,
    ) -> None:
        graph.add_node(
            topic_name,
            **{
                NodeDataFields.LABEL: topic_name,
                NodeDataFields.NODE_TYPE: NodeTypesEnum.ERROR_TOPIC,
            },
        )
        graph.add_edge(app_id, topic_name)

    def reset(self) -> None:
        self.graph.clear()
        self.json_graph.clear()
        self.pipelines.clear()
        self.json_pipelines.clear()

    @staticmethod
    def __get_json_graph(graph: nx.Graph) -> dict:
        json_graph: dict = nx.node_link_data(graph)
        json_graph["edges"] = json_graph.pop("links")
        return json_graph

    @staticmethod
    def __extract_independent_graph_components(graph: nx.Graph) -> list[nx.Graph]:
        independent_graphs = list(nx.connected_components(graph.to_undirected()))
        subgraphs: list[nx.Graph] = []
        for pipeline in independent_graphs:
            subgraph: nx.Graph = graph.subgraph(pipeline)
            subgraphs.append(subgraph)
        return subgraphs

    @staticmethod
    async def __get_positioned_json_graph(graph: nx.Graph) -> dict:
        subgraphs = DataFlowGraph.__extract_independent_graph_components(graph)
        pos = graphviz_layout(graph, prog="dot", args=settings.graph.layout_arguments)
        x = {n: p[0] for n, p in pos.items()}
        y = {n: p[1] for n, p in pos.items()}
        nx.set_node_attributes(graph, x, "x")
        nx.set_node_attributes(graph, y, "y")
        for i, subgraph in enumerate(subgraphs):
            offset = i * settings.graph.pipeline_distance
            for node_name in list(subgraph.nodes):
                graph.nodes[node_name]["y"] += offset
        return DataFlowGraph.__get_json_graph(graph)



import re
from abc import ABC, abstractmethod
from typing import Literal

from pydantic import BaseConfig, BaseModel, Extra, Field, PrivateAttr


class RouterTransformerConfig(BaseModel, ABC):
    _type: str = Field(..., alias="type")
    topics: list[str]

    @abstractmethod
    def transform_topic(self, topic: str) -> str:
        ...

    def get_routes(self) -> list[str]:
        return [self.transform_topic(topic) for topic in self.topics]


class RegexRouterTransformerConfig(RouterTransformerConfig):
    _type: Literal["org.apache.kafka.connect.transforms.RegexRouter"] = Field(
        ..., alias="type"
    )
    regex: str | None = None
    replacement: str

    _rx: re.Pattern[str] = PrivateAttr()

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        if self.regex is not None:
            self._rx = re.compile(self.regex)
        self.replacement = self.replacement.replace("$", "\\")

    def transform_topic(self, topic: str) -> str:
        return re.sub(self._rx, self.replacement, topic, 1)

    def get_routes(self) -> list[str]:
        if self.regex is not None:
            return super().get_routes()
        return [self.replacement]


class TimestampRouterTransformerConfig(RouterTransformerConfig):
    _type: Literal["org.apache.kafka.connect.transforms.TimestampRouter"] = Field(
        ..., alias="type"
    )
    topic_format: str = Field(..., alias="topic.format")
    timestamp_format: str = Field(..., alias="timestamp.format")

    def transform_topic(self, topic: str) -> str:
        transformed_topic = self.topic_format
        transformed_topic = transformed_topic.replace("${topic}", topic)
        transformed_topic = transformed_topic.replace(
            "${timestamp}", "${" + self.timestamp_format + "}"
        )
        return transformed_topic


# Used as fallback for unknown transforms
class GenericTransformerConfig(RouterTransformerConfig):
    _type: str = Field(..., alias="type")

    class Config(BaseConfig):
        extra = Extra.allow

    def transform_topic(self, topic: str) -> str:
        return topic



from typing import Any, NamedTuple

from pydantic import BaseModel, Field

from streams_explorer.models.k8s import K8sReason

GraphNode = tuple[str, dict]
GraphEdge = tuple[str, str]


class Node(BaseModel):
    id: str
    label: str
    node_type: str
    x: int | None = Field(default=None)
    y: int | None = Field(default=None)


class Metric(BaseModel):
    node_id: str
    messages_in: float | None = Field(default=None)
    messages_out: float | None = Field(default=None)
    consumer_lag: int | None = Field(default=None)
    consumer_read_rate: float | None = Field(default=None)
    topic_size: int | None = Field(default=None)
    replicas: int | None = Field(default=None)
    replicas_available: int | None = Field(default=None)
    connector_tasks: int | None = Field(default=None)


class Edge(BaseModel):
    source: str
    target: str


class Graph(BaseModel):
    directed: bool
    multigraph: bool
    graph: Any
    nodes: list[Node]
    edges: list[Edge]


class ReplicaCount(NamedTuple):
    ready: int | None
    total: int | None


class AppState(BaseModel):
    id: str
    state: K8sReason
    replicas: ReplicaCount



import httpx
from loguru import logger

from streams_explorer.core.config import settings
from streams_explorer.extractors import extractor_container
from streams_explorer.models.kafka_connector import KafkaConnector

url = settings.kafkaconnect.url
protected_keys: dict[str, set[str]] = {}


class KafkaConnect:
    @staticmethod
    def get_connectors() -> list:
        response = httpx.get(f"{url}/connectors")
        return response.json()

    @staticmethod
    def get_connector_info(connector_name: str) -> dict:
        logger.info("Get connector information for {}", connector_name)
        response = httpx.get(f"{url}/connectors/{connector_name}")
        info: dict = response.json()
        return info

    @staticmethod
    def get_connector_config(connector_name: str) -> dict:
        info = KafkaConnect.get_connector_info(connector_name)
        return KafkaConnect.sanitize_connector_config(info["config"])

    @staticmethod
    def extract_connector_class_basename(connector_class: str) -> str:
        if "." in connector_class:
            return connector_class.rsplit(".", 1)[-1]
        return connector_class

    @staticmethod
    def retrieve_connector_class_protected_keys(
        connector_class: str, config: dict
    ) -> None:
        response = httpx.put(
            f"{url}/connector-plugins/{connector_class}/config/validate",
            json=config,
        )
        if response.is_error:
            logger.warning(
                'Couldn\'t retrieve connector class validation for "{}": {}',
                connector_class,
                response,
            )
            return
        data = response.json()
        protected_keys[connector_class] = {
            config["value"]["name"]
            for config in data["configs"]
            if config["definition"]["type"] == "PASSWORD"
        }

    @staticmethod
    def sanitize_connector_config(config: dict) -> dict:
        connector_class = KafkaConnect.extract_connector_class_basename(
            config["connector.class"]
        )
        if connector_class not in protected_keys:
            KafkaConnect.retrieve_connector_class_protected_keys(
                connector_class, config
            )
        for key in protected_keys[connector_class].intersection(config):
            config[key] = "[hidden]"
            logger.debug('Sanitized connector config "{}"', key)
        return config

    @staticmethod
    def connectors() -> list[KafkaConnector]:
        protected_keys.clear()
        connectors = KafkaConnect.get_connectors()
        out = []
        for name in connectors:
            info = KafkaConnect.get_connector_info(name)
            if connector := extractor_container.on_connector_info_parsing(info, name):
                out.append(connector)
            else:
                logger.warning("Failed to parse connector {}", name)
        return out



from __future__ import annotations

from typing import TYPE_CHECKING, NamedTuple

from kubernetes_asyncio.client import V1beta1CronJob, V1Job
from loguru import logger

from streams_explorer.core.extractor.default.generic import GenericSink, GenericSource
from streams_explorer.core.extractor.extractor import (
    ConnectorExtractor,
    Extractor,
    ProducerAppExtractor,
    StreamsAppExtractor,
)
from streams_explorer.models.k8s import K8sConfig
from streams_explorer.models.kafka_connector import KafkaConnector
from streams_explorer.models.sink import Sink
from streams_explorer.models.source import Source

if TYPE_CHECKING:
    from streams_explorer.core.k8s_app import K8sAppCronJob, K8sAppJob


class SourcesSinks(NamedTuple):
    sources: list[Source]
    sinks: list[Sink]


class ExtractorContainer:
    def __init__(self, extractors: list[Extractor] | None = None) -> None:
        self.extractors: list[Extractor] = extractors if extractors else []

    def add(self, extractor: Extractor) -> None:
        self.extractors.append(extractor)
        logger.info("Added extractor {}", extractor.__class__.__name__)

    def add_generic(self) -> None:
        self.add(GenericSink())
        self.add(GenericSource())

    def reset(self) -> None:
        for extractor in self.extractors:
            extractor.reset()

    def reset_connectors(self) -> None:
        for extractor in self.extractors:
            if isinstance(extractor, ConnectorExtractor):
                extractor.reset()

    def on_streaming_app_add(self, config: K8sConfig) -> None:
        for extractor in self.extractors:
            if isinstance(extractor, StreamsAppExtractor):
                extractor.on_streaming_app_add(config)

    def on_streaming_app_delete(self, config: K8sConfig) -> None:
        for extractor in self.extractors:
            if isinstance(extractor, StreamsAppExtractor):
                extractor.on_streaming_app_delete(config)

    def on_connector_info_parsing(
        self, info: dict, connector_name: str
    ) -> KafkaConnector | None:
        for extractor in self.extractors:
            if isinstance(extractor, ConnectorExtractor):
                if connector := extractor.on_connector_info_parsing(
                    info, connector_name
                ):
                    return connector

    def on_job(self, job: V1Job) -> K8sAppJob | None:
        for extractor in self.extractors:
            if isinstance(extractor, ProducerAppExtractor):
                if app := extractor.on_job_parsing(job):
                    return app

    def on_cron_job(self, cron_job: V1beta1CronJob) -> K8sAppCronJob | None:
        for extractor in self.extractors:
            if isinstance(extractor, ProducerAppExtractor):
                if app := extractor.on_cron_job_parsing(cron_job):
                    return app

    def get_sources_sinks(self) -> SourcesSinks:
        sources: list[Source] = []
        sinks: list[Sink] = []
        for extractor in self.extractors:
            sources.extend(extractor.sources)
            sinks.extend(extractor.sinks)
        return SourcesSinks(sources, sinks)



from fastapi import APIRouter, Depends
from loguru import logger

from streams_explorer.api.dependencies.streams_explorer import get_streams_explorer
from streams_explorer.streams_explorer import StreamsExplorer

router = APIRouter()


@router.post("")
async def update(
    streams_explorer: StreamsExplorer = Depends(get_streams_explorer),
):
    logger.info("(Re-)Building graphs")
    await streams_explorer.update_graph()



from loguru import logger

from streams_explorer.core.services.metric_providers import (
    MetricProvider,
    PrometheusMetricProvider,
)
from streams_explorer.plugins import load_plugin


def load_metric_provider() -> type[MetricProvider]:
    metric_provider = load_plugin(MetricProvider)
    if not metric_provider:
        logger.info("Using default PrometheusMetricProvider")
        return PrometheusMetricProvider
    logger.info(f"Using custom MetricProvider: {metric_provider.__name__}")
    return metric_provider



from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

from kubernetes_asyncio.client import V1Container

from streams_explorer.models.k8s import K8sConfig
from streams_explorer.plugins import Plugin

if TYPE_CHECKING:
    from streams_explorer.core.k8s_app import K8sApp


class K8sConfigParser(Plugin):
    """Base class for parsing configuration of streaming application deployments."""

    def __init__(self, k8s_app: K8sApp) -> None:
        self.k8s_app = k8s_app

    @abstractmethod
    def parse(self) -> K8sConfig:
        ...


class StreamsBootstrapConfigParser(K8sConfigParser):
    """Config parser for deployments configured through streams-bootstrap."""

    def __init__(self, k8s_app: K8sApp) -> None:
        super().__init__(k8s_app)
        self._id = self.get_id()
        self.config = K8sConfig(self._id, name=self.get_name())

    def get_id(self) -> str:
        name: str | None = None
        if self.k8s_app.metadata.labels:
            name = self.k8s_app.metadata.labels.get("app")
        if not name:
            name = self.k8s_app.metadata.name
        if not name:
            raise TypeError(f"Name is required for {self.k8s_app.class_name}")
        return name

    def get_name(self) -> str:
        return self._id

    def parse_config(self, name: str, value: str) -> None:
        match name:
            case "INPUT_TOPICS":
                self.config.input_topics = self.parse_input_topics(value)
            case "OUTPUT_TOPIC":
                self.config.output_topic = value
            case "ERROR_TOPIC":
                self.config.error_topic = value
            case "EXTRA_INPUT_TOPICS":
                self.config.extra_input_topics = self.parse_extra_topics(value)
            case "EXTRA_OUTPUT_TOPICS":
                self.config.extra_output_topics = self.parse_extra_topics(value)
            case "INPUT_PATTERN":
                self.config.input_pattern = value
            case "EXTRA_INPUT_PATTERNS":
                self.config.extra_input_patterns = self.parse_extra_topics(value)
            case _:
                self.config.extra[name] = value

    @staticmethod
    def parse_input_topics(input_topics: str) -> list[str]:
        return input_topics.split(",")

    @staticmethod
    def parse_extra_topics(extra_topics: str) -> list[str]:
        extra_topics = extra_topics.removesuffix(",")  # remove trailing comma
        return [
            topic
            for role in extra_topics.split(",")
            for topic in role.split("=")[1].split(";")
        ]


class StreamsBootstrapEnvParser(StreamsBootstrapConfigParser):
    """Default parser for streams-bootstrap deployments configured through environment variables."""

    def __init__(self, k8s_app: K8sApp) -> None:
        super().__init__(k8s_app)
        self.env_prefix = self.get_env_prefix(self.k8s_app.container)

    def parse(self) -> K8sConfig:
        container = self.k8s_app.container

        if not container or not container.env:
            return self.config

        for env in container.env:
            if env.value is not None:
                name = self.__normalise_name(env.name)
                self.parse_config(name, env.value)
        return self.config

    def __normalise_name(self, name: str) -> str:
        if self.env_prefix:
            return name.removeprefix(self.env_prefix)
        return name

    @staticmethod
    def get_env_prefix(container: V1Container | None) -> str | None:
        if container and container.env:
            for env_var in container.env:
                if env_var.name == "ENV_PREFIX":
                    return env_var.value


class StreamsBootstrapArgsParser(StreamsBootstrapConfigParser):
    """Optional parser for streams-bootstrap deployments configured through CLI arguments."""

    def parse(self) -> K8sConfig:
        container = self.k8s_app.container

        if not container or not container.args:
            return self.config

        args: list[str] = container.args

        for arg in args:
            arg = arg.removeprefix("--")
            name, value = arg.split("=")
            if not name or not value:
                continue
            name = self.__normalise_name(name)
            self.parse_config(name, value)
        return self.config

    @staticmethod
    def __normalise_name(name: str) -> str:
        return name.upper().replace("-", "_")



from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum


@dataclass
class K8sConfig:
    id: str
    name: str
    input_topics: list[str] = field(default_factory=list)  # required for streaming app
    output_topic: str | None = None  # required for streaming app
    error_topic: str | None = None
    input_pattern: str | None = None
    extra_input_topics: list[str] = field(default_factory=list)
    extra_output_topics: list[str] = field(default_factory=list)
    extra_input_patterns: list[str] = field(default_factory=list)
    extra: dict[str, str] = field(default_factory=dict)


class K8sDeploymentUpdateType(str, Enum):
    ADDED = "ADDED"
    MODIFIED = "MODIFIED"
    DELETED = "DELETED"


# Documentation: https://github.com/tomplus/kubernetes_asyncio/blob/11c3eb4d50ae822545572aa7b8c15f7153f65a1c/kubernetes_asyncio/docs/EventsV1Event.md
class K8sEventType(str, Enum):
    NORMAL = "Normal"
    WARNING = "Warning"


class K8sReason(str, Enum):
    UNKNOWN = "Unknown"

    # Container event reason list
    CREATED = "Created"
    STARTED = "Started"
    FAILED = "Failed"
    KILLING = "Killing"
    PREEMPTING = "Preempting"
    BACKOFF = "BackOff"
    EXCEEDEDGRACEPERIOD = "ExceededGracePeriod"

    # Image event reason list
    PULLING = "Pulling"
    PULLED = "Pulled"

    @staticmethod
    def from_str(reason: str) -> K8sReason:
        try:
            return K8sReason[reason.upper()]
        except KeyError:
            raise NotImplementedError(f"Kubernetes event reason '{reason}' is unknown")



from streams_explorer.core.extractor.extractor import ConnectorExtractor
from streams_explorer.models.kafka_connector import (
    KafkaConnector,
    KafkaConnectorTypesEnum,
)


class GenericSink(ConnectorExtractor):
    def on_connector_info_parsing(
        self, info: dict, connector_name: str
    ) -> KafkaConnector | None:
        if info["type"] == KafkaConnectorTypesEnum.SINK:
            config = info["config"]
            return KafkaConnector(
                name=connector_name,
                config=config,
                type=KafkaConnectorTypesEnum.SINK,
            )


class GenericSourceConnector(KafkaConnector):
    def __init__(self, **kwargs) -> None:
        super().__init__(
            **kwargs,
            type=KafkaConnectorTypesEnum.SOURCE,
        )

    def get_topics(self) -> list[str]:
        return []


class GenericSource(ConnectorExtractor):
    def on_connector_info_parsing(
        self, info: dict, connector_name: str
    ) -> KafkaConnector | None:
        if info["type"] == KafkaConnectorTypesEnum.SOURCE:
            return GenericSourceConnector(name=connector_name, config=info["config"])



from fastapi import FastAPI
from starlette.staticfiles import StaticFiles

from streams_explorer.api.routes import api
from streams_explorer.core.config import API_PREFIX, APP_NAME


def get_application() -> FastAPI:
    app = FastAPI(title=APP_NAME)
    app.include_router(api.router, prefix=API_PREFIX)

    app.mount("/", StaticFiles(directory="static", html=True), name="static")

    return app



from enum import Enum
from typing import Union

import pydantic
from loguru import logger
from pydantic import BaseConfig, BaseModel, Extra, Field, ValidationError

from streams_explorer.core.extractor.default.transformer import (
    GenericTransformerConfig,
    RegexRouterTransformerConfig,
    RouterTransformerConfig,
    TimestampRouterTransformerConfig,
)


class KafkaConnectorTypesEnum(str, Enum):
    SINK = "sink"
    SOURCE = "source"


class KafkaConnectorConfig(BaseModel):
    topics: str | None = Field(default=None)
    transforms: str | None = Field(default=None)
    error_topic: str | None = Field(
        default=None, alias="errors.deadletterqueue.topic.name"
    )

    class Config(BaseConfig):
        extra = Extra.allow

    def get_transforms(self) -> list[str]:
        if self.transforms is not None:
            return self.transforms.split(",")
        return []


class KafkaConnector(BaseModel):
    name: str
    type: KafkaConnectorTypesEnum
    config: KafkaConnectorConfig
    error_topic: str | None = Field(default=None)
    topics: list[str] | None = Field(
        default=None
    )  # Deprecated please override get_topics for your kafka connector.

    # Do not change order here. The fallback GenericTransformerConfig should always be the last element
    _transformers = Union[
        RegexRouterTransformerConfig,
        TimestampRouterTransformerConfig,
        GenericTransformerConfig,
    ]

    def get_topics(self) -> list[str]:
        """
        Override in your kafka connector. Use the config to retrieve and parse the topics.
        This implementation only ensures the support of (older) plugins that use the deprecated topics field.
        """
        if self.topics is None:
            return KafkaConnector.split_topics(self.config.topics)
        return self.topics

    def get_error_topic(self):
        if self.error_topic is None:
            return self.config.error_topic
        return self.error_topic

    @staticmethod
    def split_topics(topics: str | None) -> list[str]:
        if topics:
            return topics.replace(" ", "").split(",")
        return []

    def get_routes(self) -> set[str]:
        """
        Applies the single message transformers to get target routes (e.g. for the ElasticSearchSinkConnector)
        """
        routes = self.get_topics()
        transforms = self.config.get_transforms()
        for transformer_name in transforms:
            routes = self.apply_transformer(routes, transformer_name)
        return set(routes)

    def apply_transformer(self, indices: list[str], transformer_name: str) -> list[str]:
        transformer_prefix = KafkaConnector.get_transformer_prefix(transformer_name)
        # transformer config without transformer_prefix
        transformer_config = {
            config_name.replace(transformer_prefix, ""): config_value
            for config_name, config_value in self.config.dict().items()
            if KafkaConnector.is_transformer_config(config_name, transformer_prefix)
        }
        transformer_config["topics"] = indices
        try:
            transformer: RouterTransformerConfig = pydantic.parse_obj_as(
                # HACK: https://github.com/samuelcolvin/pydantic/issues/1847
                self._transformers,  # type: ignore[arg-type]
                transformer_config,
            )
            return transformer.get_routes()
        except ValidationError as e:
            logger.exception(
                f"Failed to apply transformer {transformer_name} on {self.name}", e
            )
            return []

    @staticmethod
    def get_transformer_prefix(transformer_name: str) -> str:
        return f"transforms.{transformer_name}."

    @staticmethod
    def is_transformer_config(key: str, transformer_prefix: str) -> bool:
        return key.startswith(transformer_prefix)



from streams_explorer.core.extractor.extractor import ConnectorExtractor
from streams_explorer.models.kafka_connector import (
    KafkaConnector,
    KafkaConnectorTypesEnum,
)
from streams_explorer.models.sink import Sink


class ElasticsearchSink(ConnectorExtractor):
    def on_connector_info_parsing(
        self, info: dict, connector_name: str
    ) -> KafkaConnector | None:
        config = info["config"]
        connector_class: str | None = config.get("connector.class")
        if connector_class and "ElasticsearchSinkConnector" in connector_class:

            connector = KafkaConnector(
                name=connector_name,
                config=config,
                type=KafkaConnectorTypesEnum.SINK,
            )
            indices = connector.get_routes()
            for index in indices:
                self.sinks.append(
                    Sink(
                        name=index,
                        node_type="elasticsearch-index",
                        source=connector_name,
                    )
                )
            return connector



from __future__ import annotations

from typing import TypeAlias

from kubernetes_asyncio.client import (
    V1beta1CronJob,
    V1Container,
    V1Deployment,
    V1Job,
    V1ObjectMeta,
    V1PodSpec,
    V1StatefulSet,
)
from loguru import logger

from streams_explorer.core.config import settings
from streams_explorer.core.k8s_config_parser import K8sConfigParser
from streams_explorer.k8s_config_parser import load_config_parser
from streams_explorer.models.graph import AppState, ReplicaCount
from streams_explorer.models.k8s import K8sConfig, K8sReason

ATTR_PIPELINE = "pipeline"

K8sObject: TypeAlias = V1Deployment | V1StatefulSet | V1Job | V1beta1CronJob

config_parser: type[K8sConfigParser] = load_config_parser()


class K8sApp:
    def __init__(self, k8s_object: K8sObject) -> None:
        self.k8s_object = k8s_object
        self.metadata: V1ObjectMeta = k8s_object.metadata or V1ObjectMeta()
        self.attributes: dict[str, str] = {}
        self.config: K8sConfig
        self.state: K8sReason = K8sReason.UNKNOWN
        self.setup()

    @property
    def id(self) -> str:
        return self.config.id

    @property
    def name(self) -> str:
        return self.config.name

    @property
    def input_topics(self) -> list[str]:
        return self.config.input_topics

    @property
    def output_topic(self) -> str | None:
        return self.config.output_topic

    @property
    def error_topic(self) -> str | None:
        return self.config.error_topic

    @property
    def input_pattern(self) -> str | None:
        return self.config.input_pattern

    @property
    def extra_input_topics(self) -> list[str]:
        return self.config.extra_input_topics

    @property
    def extra_output_topics(self) -> list[str]:
        return self.config.extra_output_topics

    @property
    def extra_input_patterns(self) -> list[str]:
        return self.config.extra_input_patterns

    @property
    def pipeline(self) -> str | None:
        return self.attributes.get(settings.k8s.pipeline.label)

    @property
    def consumer_group(self) -> str | None:
        return self.attributes.get(settings.k8s.consumer_group_annotation)

    @property
    def replicas_ready(self) -> int | None:
        if not self.k8s_object.status:
            return None
        return self.k8s_object.status.ready_replicas

    @property
    def replicas_total(self) -> int | None:
        if not self.k8s_object.status:
            return None
        return self.k8s_object.status.replicas

    def setup(self) -> None:
        self.spec = self._get_pod_spec()
        self._ignore_containers = self.get_ignore_containers()
        self.container = self.get_app_container(self.spec, self._ignore_containers)
        self.extract_config()
        self.__set_attributes()

    def _get_pod_spec(self) -> V1PodSpec | None:
        if self.k8s_object.spec and self.k8s_object.spec.template.spec:
            return self.k8s_object.spec.template.spec

    def extract_config(self) -> None:
        self.config = config_parser(self).parse()

    def is_streams_app(self) -> bool:
        if not self.input_topics and not self.output_topic:
            return False
        return True

    def to_dict(self) -> dict:
        return self.k8s_object.to_dict()

    def to_state_update(self) -> AppState:
        return AppState(
            id=self.id,
            replicas=ReplicaCount(self.replicas_ready, self.replicas_total),
            state=self.state,
        )

    @property
    def class_name(self) -> str:
        return self.__class__.__name__

    def __set_attributes(self) -> None:
        self._set_labels()
        self._set_pipeline()
        self._set_annotations()

    def _set_labels(self) -> None:
        labels = self.metadata.labels
        if not labels:
            return

        for key in self._labels_to_use():
            value = labels.get(key)
            if value is not None:
                self.attributes[key] = value
            elif self.is_streams_app():
                logger.warning(
                    f"{self.class_name} {self.name} does not have a label with the name: {key}"
                )

    def _set_pipeline(self) -> None:
        if self.pipeline:
            self.attributes[ATTR_PIPELINE] = self.pipeline

    def _set_annotations(self) -> None:
        if (
            self.k8s_object.spec
            and self.k8s_object.spec.template.metadata
            and self.k8s_object.spec.template.metadata.annotations
        ):
            annotations = self.k8s_object.spec.template.metadata.annotations
            self.attributes.update(annotations)

    @staticmethod
    def factory(k8s_object: K8sObject) -> K8sApp:
        match k8s_object:
            case V1Deployment():  # type: ignore[misc]
                return K8sAppDeployment(k8s_object)
            case V1StatefulSet():  # type: ignore[misc]
                return K8sAppStatefulSet(k8s_object)
            case V1Job():  # type: ignore[misc]
                return K8sAppJob(k8s_object)
            case V1beta1CronJob():  # type: ignore[misc]
                return K8sAppCronJob(k8s_object)
            case _:
                raise ValueError(k8s_object)

    @staticmethod
    def get_app_container(
        spec: V1PodSpec | None, ignore_containers: set[str] = set()
    ) -> V1Container | None:
        if spec and spec.containers:
            for container in spec.containers:
                if container.name not in ignore_containers:
                    return container

    @staticmethod
    def get_ignore_containers() -> set[str]:
        return {container["name"] for container in settings.k8s.containers.ignore}

    @staticmethod
    def _labels_to_use() -> set[str]:
        return set(settings.k8s.labels)


class K8sAppJob(K8sApp):
    def __init__(self, k8s_object: V1Job) -> None:
        super().__init__(k8s_object)

    def setup(self) -> None:
        self.spec = self._get_pod_spec()
        self.container = self.get_app_container(self.spec)
        self.extract_config()
        self.__set_attributes()

    @property
    def replicas_ready(self) -> None:
        return None

    @property
    def replicas_total(self) -> None:
        return None

    def _get_pod_spec(self) -> V1PodSpec | None:
        if self.k8s_object.spec and self.k8s_object.spec.template.spec:
            return self.k8s_object.spec.template.spec

    def __set_attributes(self) -> None:
        self._set_labels()
        self._set_pipeline()


class K8sAppCronJob(K8sApp):
    def __init__(self, k8s_object: V1beta1CronJob) -> None:
        super().__init__(k8s_object)

    def setup(self) -> None:
        self.spec = self._get_pod_spec()
        self.container = self.get_app_container(self.spec)
        self.extract_config()
        self.__set_attributes()

    @property
    def replicas_ready(self) -> None:
        return None

    @property
    def replicas_total(self) -> None:
        return None

    def _get_pod_spec(self) -> V1PodSpec | None:
        if (
            self.k8s_object.spec
            and self.k8s_object.spec.job_template.spec
            and self.k8s_object.spec.job_template.spec.template.spec
        ):
            return self.k8s_object.spec.job_template.spec.template.spec

    def __set_attributes(self) -> None:
        self._set_labels()
        self._set_pipeline()


class K8sAppDeployment(K8sApp):
    def __init__(self, k8s_object: V1Deployment) -> None:
        super().__init__(k8s_object)


class K8sAppStatefulSet(K8sApp):
    def __init__(self, k8s_object: V1StatefulSet) -> None:
        super().__init__(k8s_object)

    def get_service_name(self) -> str | None:
        if self.k8s_object.spec:
            return self.k8s_object.spec.service_name



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['streams_explorer',
 'streams_explorer.api',
 'streams_explorer.api.dependencies',
 'streams_explorer.api.routes',
 'streams_explorer.core',
 'streams_explorer.core.extractor',
 'streams_explorer.core.extractor.default',
 'streams_explorer.core.services',
 'streams_explorer.models']

package_data = \
{'': ['*']}

install_requires = \
['cachetools>=4.2.2,<5.0.0',
 'confluent-kafka>=1.9.2,<2.0.0',
 'dynaconf>=3.1.7,<4.0.0',
 'fastapi-utils>=0.2.1,<0.3.0',
 'fastapi>=0.78.0,<0.79.0',
 'httpx>=0.23.1,<0.24.0',
 'kubernetes-asyncio>=23.6.0,<24.0.0',
 'loguru>=0.5.3,<0.6.0',
 'matplotlib>=3.4.3,<4.0.0',
 'networkx>=2.6.3,<3.0.0',
 'pydantic>=1.9.1,<2.0.0',
 'pygraphviz>=1.10,<2.0',
 'uvicorn[standard]>=0.20.0,<0.21.0']

entry_points = \
{'console_scripts': ['start = main:start']}

setup_kwargs = {
    'name': 'streams-explorer',
    'version': '2.2.1',
    'description': 'Explore Data Pipelines in Apache Kafka.',
    'long_description': '# Streams Explorer\n\n> Explore Data Pipelines in Apache Kafka.\n\n![streams-explorer](https://github.com/bakdata/streams-explorer/blob/main/screens/overview.png?raw=true)\n\n> **Note**\n> We are participating in the annual Hacktoberfest. If you\'re looking to contribute, please see our [open issues](https://github.com/bakdata/streams-explorer/issues?q=is%3Aissue+is%3Aopen+sort%3Aupdated-desc+label%3Ahacktoberfest) and use the [standalone installation](#standalone) for development.\n\n## Contents\n\n- [Streams Explorer](#streams-explorer)\n  - [Features](#features)\n  - [Overview](#overview)\n  - [Installation](#installation)\n    - [Docker Compose](#docker-compose)\n    - [Deploying to Kubernetes cluster](#deploying-to-kubernetes-cluster)\n    - [Standalone](#standalone)\n      - [Backend](#backend)\n      - [Frontend](#frontend)\n  - [Configuration](#configuration)\n    - [Kafka](#kafka)\n    - [Kafka Connect](#kafka-connect)\n    - [Kubernetes](#kubernetes)\n    - [Schema Registry / Karapace](#schema-registry--karapace)\n    - [Prometheus](#prometheus)\n    - [AKHQ](#akhq)\n    - [Redpanda Console](#redpanda-console)\n    - [Grafana](#grafana)\n    - [Kibana](#kibana)\n    - [Elasticsearch](#elasticsearch)\n    - [Plugins](#plugins)\n  - [Demo pipeline](#demo-pipeline)\n  - [Plugin customization](#plugin-customization)\n\n## Features\n\n- Visualization of streaming applications, topics, and connectors\n- Monitor all or individual pipelines from multiple namespaces\n- Inspection of Avro schema from schema registry\n- Integration with [streams-bootstrap](https://github.com/bakdata/streams-bootstrap) and [faust-bootstrap](https://github.com/bakdata/faust-bootstrap), or custom streaming app config parsing from Kubernetes deployments using plugins\n- Real-time metrics from Prometheus (consumer lag & read rate, replicas, topic size, messages in & out per second, connector tasks)\n- Linking to external services for logging and analysis, such as Kibana, Grafana, Loki, AKHQ, Redpanda Console, and Elasticsearch\n- Customizable through Python plugins\n\n## Overview\n\nVisit our introduction [blogpost](https://medium.com/bakdata/exploring-data-pipelines-in-apache-kafka-with-streams-explorer-8337dd11fdad) for a complete overview and demo of Streams Explorer.\n\n## Installation\n\n### Docker Compose\n\n1. Forward the ports to Prometheus. (Kafka Connect, Schema Registry, and other integrations are optional)\n2. Start the container\n\n```sh\ndocker compose up\n```\n\nOnce the container is started visit <http://localhost:8080>\n\n### Deploying to Kubernetes cluster\n\n1. Add the Helm chart repository\n\n```sh\nhelm repo add streams-explorer https://bakdata.github.io/streams-explorer\n```\n\n2. Install\n\n```sh\nhelm upgrade --install --values helm-chart/values.yaml streams-explorer streams-explorer/streams-explorer\n```\n\n### Standalone\n\n#### Backend\n\n1. Install dependencies using [Poetry](https://python-poetry.org)\n\n```sh\npoetry install\n```\n\n2. Forward the ports to Prometheus. (Kafka Connect, Schema Registry, and other integrations are optional)\n3. Configure the backend in [settings.yaml](backend/settings.yaml).\n4. Start the backend server\n\n```sh\npoetry run start\n```\n\n#### Frontend\n\n1. Install dependencies\n\n```sh\nnpm ci\n```\n\n2. Start the frontend server\n\n```sh\nnpm run build && npm run prod\n```\n\nVisit <http://localhost:3000>\n\n## Configuration\n\nDepending on your type of installation set the configuration for the backend server in this file:\n\n- **Docker Compose**: [docker-compose.yaml](docker-compose.yaml)\n- **Kubernetes**: [helm-chart/values.yaml](helm-chart/values.yaml)\n- **standalone**: [backend/settings.yaml](backend/settings.yaml)\n\nIn the [helm-chart/values.yaml](helm-chart/values.yaml) configuration is done either through the `config` section using double underscore notation, e.g. `K8S__consumer_group_annotation: consumerGroup` or the content of [backend/settings.yaml](backend/settings.yaml) can be pasted under the `settings` section. Alternatively all configuration options can be written as environment variables using double underscore notation and the prefix `SE`, e.g. `SE_K8S__deployment__cluster=false`.\n\nThe following configuration options are available:\n\n#### General\n\n- `graph.update_interval` Render the graph every x seconds (int, **required**, default: `300`)\n- `graph.layout_arguments` Arguments passed to graphviz layout (string, **required**, default: `-Grankdir=LR -Gnodesep=0.8 -Gpad=10`)\n- `graph.pipeline_distance` Increase/decrease vertical space between pipeline graphs by X pixels (int, **required**, default: `500`)\n- `graph.resolve.input_pattern_topics.all` If true topics that match (extra) input pattern(s) are connected to the streaming app in the graph containing all pipelines (bool, **required**, default: `false`)\n- `graph.resolve.input_pattern_topics.pipelines` If true topics that match (extra) input pattern(s) are connected to the streaming app in pipeline graphs (bool, **required**, default: `false`)\n\n#### Kafka\n\n- `kafka.enable` Enable Kafka (bool, default: `false`)\n- `kafka.config` librdkafka configuration properties ([reference](https://github.com/edenhill/librdkafka/blob/master/CONFIGURATION.md)) (dict, default: `{"bootstrap.servers": "localhost:9092"}`)\n- `kafka.displayed_information` Configuration options of Kafka topics displayed in the frontend (list of dict)\n- `kafka.topic_names_cache.ttl` Cache for retrieving all topic names (used when input topic patterns are resolved) (int, default: `3600`)\n\n#### Kafka Connect\n\n- `kafkaconnect.url` URL of Kafka Connect server (string, default: None)\n- `kafkaconnect.update_interval` Fetch connectors every x seconds (int, default: `300`)\n- `kafkaconnect.displayed_information` Configuration options of Kafka connectors displayed in the frontend (list of dict)\n\n#### Kubernetes\n\n- `k8s.deployment.cluster` Whether streams-explorer is deployed to Kubernetes cluster (bool, **required**, default: `false`)\n- `k8s.deployment.context` Name of cluster (string, optional if running in cluster, default: `kubernetes-cluster`)\n- `k8s.deployment.namespaces` Kubernetes namespaces (list of string, **required**, default: `[\'kubernetes-namespace\']`)\n- `k8s.containers.ignore` Name of containers that should be ignored/hidden (list of string, default: `[\'prometheus-jmx-exporter\']`)\n- `k8s.displayed_information` Details of pod that should be displayed (list of dict, default: `[{\'name\': \'Labels\', \'key\': \'metadata.labels\'}]`)\n- `k8s.labels` Labels used to set attributes of nodes (list of string, **required**, default: `[\'pipeline\']`)\n- `k8s.pipeline.label` Attribute of nodes the pipeline name should be extracted from (string, **required**, default: `pipeline`)\n- `k8s.consumer_group_annotation` Annotation the consumer group name should be extracted from (string, **required**, default: `consumerGroup`)\n\n#### Schema Registry / Karapace\n\n- `schemaregistry.url` URL of Confluent Schema Registry or Karapace (string, default: None)\n\n#### Prometheus\n\n- `prometheus.url` URL of Prometheus (string, **required**, default: `http://localhost:9090`)\n\nThe following exporters are required to collect Kafka metrics for Prometheus:\n\n- [Kafka Exporter](https://github.com/danielqsj/kafka_exporter)\n- [Kafka Lag Exporter](https://github.com/lightbend/kafka-lag-exporter)\n- [Kafka Connect Exporter](https://github.com/wakeful/kafka_connect_exporter)\n\n#### AKHQ\n\n- `akhq.enable` Enable AKHQ (bool, default: `false`)\n- `akhq.url` URL of AKHQ (string, default: `http://localhost:8080`)\n- `akhq.cluster` Name of cluster (string, default: `kubernetes-cluster`)\n- `akhq.connect` Name of connect (string, default: None)\n\n#### Redpanda Console\n\nRedpanda Console can be used instead of AKHQ. (mutually exclusive)\n\n- `redpanda_console.enable` Enable Redpanda Console (bool, default: `false`)\n- `redpanda_console.url` URL of Redpanda Console (string, default: `http://localhost:8080`)\n\n#### Grafana\n\n- `grafana.enable` Enable Grafana (bool, default: `false`)\n- `grafana.url` URL of Grafana (string, default: `http://localhost:3000`)\n- `grafana.dashboards.topics` Path to topics dashboard (string), sample dashboards for topics and consumer groups are included in the [`./grafana`](https://github.com/bakdata/streams-explorer/tree/main/grafana) subfolder\n- `grafana.dashboards.consumergroups` Path to consumer groups dashboard (string)\n\n#### Kibana\n\n- `kibanalogs.enable` Enable Kibana logs (bool, default: `false`)\n- `kibanalogs.url` URL of Kibana logs (string, default: `http://localhost:5601`)\n\n#### Loki\n\nLoki can be used instead of Kibana. (mutually exclusive)\n\n- `loki.enable` Enable Loki logs (bool, default: `false`)\n- `loki.url` URL of Loki logs (string, default: `http://localhost:3000`)\n\n#### Elasticsearch\n\nfor Kafka Connect Elasticsearch connector\n\n- `esindex.url` URL of Elasticsearch index (string, default: `http://localhost:5601/app/kibana#/dev_tools/console`)\n\n#### Plugins\n\n- `plugins.path` Path to folder containing plugins relative to backend (string, **required**, default: `./plugins`)\n- `plugins.extractors.default` Whether to load default extractors (bool, **required**, default: `true`)\n\n## Demo pipeline\n\n![demo-pipeline](https://github.com/bakdata/streams-explorer/blob/main/screens/demo-pipeline.png?raw=true)\n\n[ATM Fraud detection with streams-bootstrap](https://github.com/bakdata/streams-explorer/blob/main/demo-atm-fraud/README.md)\n\n## Plugin customization\n\nIt is possible to create your own config parser, linker, metric provider, and extractors in Python by implementing the `K8sConfigParser`, `LinkingService`, `MetricProvider`, or `Extractor` classes. This way you can customize it to your specific setup and services. As an example we provide the [`DefaultLinker`](https://github.com/bakdata/streams-explorer/blob/main/backend/streams_explorer/defaultlinker.py) as `LinkingService`. The default [`MetricProvider`](https://github.com/bakdata/streams-explorer/blob/main/backend/streams_explorer/core/services/metric_providers.py) supports Prometheus. Furthermore the following default `Extractor` plugins are included:\n\n- [`ElasticsearchSink`](https://github.com/bakdata/streams-explorer/blob/main/backend/streams_explorer/core/extractor/default/elasticsearch_sink.py)\n- [`JdbcSink`](https://github.com/bakdata/streams-explorer/blob/main/backend/streams_explorer/core/extractor/default/jdbc_sink.py)\n- [`S3Sink`](https://github.com/bakdata/streams-explorer/blob/main/backend/streams_explorer/core/extractor/default/s3_sink.py)\n- [`GenericSink`/`GenericSource`](https://github.com/bakdata/streams-explorer/blob/main/backend/streams_explorer/core/extractor/default/generic.py)\n\nIf your streaming application deployments are configured through environment variables, following the schema of [streams-bootstrap](https://github.com/bakdata/streams-bootstrap) or [faust-bootstrap](https://github.com/bakdata/faust-bootstrap), the Streams Explorer works out-of-the-box with the default deployment parser.\nFor streams-bootstrap deployments configured through CLI arguments a separate parser can be loaded by creating a Python file (e.g. `config_parser.py`) in the plugins folder with the following import statement:\n\n```python\nfrom streams_explorer.core.k8s_config_parser import StreamsBootstrapArgsParser\n```\n\nFor other setups a custom config parser plugin can be created by inheriting from the [`K8sConfigParser`](https://github.com/bakdata/streams-explorer/blob/main/backend/streams_explorer/core/k8s_config_parser.py) class and implementing the `parse` method. In this example we\'re retrieving the streaming app configurations from an external REST API. In order for a deployment to be indentified as streaming app, input and output topics are required.\n\n```python\nimport httpx\n\nfrom streams_explorer.core.k8s_config_parser import K8sConfigParser\nfrom streams_explorer.models.k8s import K8sConfig\n\n\nclass CustomConfigParser(K8sConfigParser):\n    def get_name(self) -> str:\n        name = self.k8s_app.metadata.name\n        if not name:\n            raise TypeError(f"Name is required for {self.k8s_app.class_name}")\n        return name\n\n    def parse(self) -> K8sConfig:\n        """Retrieve app config from REST endpoint."""\n        name = self.get_name()\n        data = httpx.get(f"url/config/{name}").json()\n        return K8sConfig(**data)\n```\n',
    'author': 'bakdata',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/bakdata/streams-explorer',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import asyncio
from datetime import datetime, timedelta
from enum import Enum
from typing import Callable

import httpx
from loguru import logger

from streams_explorer.core.config import settings
from streams_explorer.models.graph import GraphNode, Metric
from streams_explorer.models.node_types import NodeTypesEnum
from streams_explorer.plugins import Plugin


class PrometheusMetric(Enum):
    def __init__(
        self,
        metric: str,
        query: str,
        key: str,
        value_transformer: Callable[[str], float],
    ) -> None:
        self.metric: str = metric
        self.query: str = query
        self._k: str = key
        self._v: Callable[[str], float] = value_transformer

    MESSAGES_IN = (
        "messages_in",
        "sum by(topic) (rate(kafka_topic_partition_current_offset[5m]))",
        "topic",
        lambda d: round(float(d), 2),
    )
    MESSAGES_OUT = (
        "messages_out",
        "sum by(topic) (rate(kafka_consumergroup_group_offset[5m]) >= 0)",
        "topic",
        lambda d: round(float(d), 2),
    )
    CONSUMER_LAG = (
        "consumer_lag",
        'sum by(group) (kafka_consumergroup_group_topic_sum_lag{group=~".+"})',
        "group",
        int,
    )
    CONSUMER_READ_RATE = (
        "consumer_read_rate",
        'sum by(group) (rate(kafka_consumergroup_group_offset{group=~".+"}[5m]) >= 0)',
        "group",
        float,
    )
    TOPIC_SIZE = (
        "topic_size",
        "sum by(topic) (kafka_topic_partition_current_offset - kafka_topic_partition_oldest_offset)",
        "topic",
        float,
    )
    REPLICAS = (
        "replicas",
        "sum by(deployment) (kube_deployment_status_replicas)",
        "deployment",
        int,
    )
    REPLICAS_AVAILABLE = (
        "replicas_available",
        "sum by(deployment) (kube_deployment_status_replicas_available)",
        "deployment",
        int,
    )
    CONNECTOR_TASKS = (
        "connector_tasks",
        "sum by(connector) (kafka_connect_connector_tasks_state == 1) or clamp_max(sum by(connector) (kafka_connect_connector_tasks_state), 0)",
        "connector",
        int,
    )

    def transform(self, data: list) -> dict:
        return {d["metric"][self._k]: self._v(d["value"][-1]) for d in data}


def sort_topics_first(nodes: list[GraphNode]) -> list[GraphNode]:
    return sorted(nodes, key=is_topic, reverse=True)


def is_topic(node: GraphNode) -> bool:
    node_type: NodeTypesEnum = node[1]["node_type"]
    return node_type in (NodeTypesEnum.TOPIC, NodeTypesEnum.ERROR_TOPIC)


class MetricProvider(Plugin):
    def __init__(self, nodes: list[GraphNode]) -> None:
        self._nodes: list[GraphNode] = sort_topics_first(nodes)
        self._metrics: list[Metric] = []
        self._data: dict[str, dict] = {}
        self._last_refresh: datetime = datetime.min
        self._cache_ttl: timedelta = timedelta(0)

    async def refresh_data(self) -> None:
        pass

    @staticmethod
    def get_consumer_group(node_id: str, node: dict) -> str | None:
        node_type: NodeTypesEnum = node["node_type"]
        if node_type == NodeTypesEnum.CONNECTOR:
            return f"connect-{node_id}"
        return node.get(settings.k8s.consumer_group_annotation)

    async def update(self) -> None:
        await self.refresh_data()
        self._metrics = [
            Metric(
                node_id=node_id,
                consumer_lag=self._data["consumer_lag"].get(
                    self.get_consumer_group(node_id, node)
                ),
                consumer_read_rate=self._data["consumer_read_rate"].get(
                    self.get_consumer_group(node_id, node)
                ),
                messages_in=self._data["messages_in"].get(node_id),
                messages_out=self._data["messages_out"].get(node_id),
                topic_size=self._data["topic_size"].get(node_id),
                replicas=self._data["replicas"].get(node_id),
                replicas_available=self._data["replicas_available"].get(node_id),
                connector_tasks=self._data["connector_tasks"].get(node_id),
            )
            for node_id, node in self._nodes
        ]

    async def get(self) -> list[Metric]:
        now = datetime.utcnow()
        cache_age = now - self._last_refresh
        if cache_age > self._cache_ttl:
            self._last_refresh = now
            await self.update()
        else:
            logger.debug("Serving cached metrics (age {}s)", cache_age.seconds)
        return self._metrics


class PrometheusException(Exception):
    pass


class PrometheusMetricProvider(MetricProvider):
    def __init__(self, nodes: list[GraphNode]) -> None:
        super().__init__(nodes)
        self._client = httpx.AsyncClient(
            base_url=f"{settings.prometheus.url}/api/v1",
        )
        # min refresh interval (set by the frontend) is 10s, intermediate requests should be cached
        self._cache_ttl = timedelta(seconds=9)

    async def _pull_metric(self, metric: PrometheusMetric) -> list:
        try:
            return await self._query(metric.query)
        except PrometheusException:
            logger.error("Error pulling {}", metric)
            return []

    async def _query(self, query: str) -> list:
        try:
            r = await self._client.get("/query", params={"query": query})
            r.raise_for_status()
            data: dict = r.json()
            return data["data"]["result"]
        except KeyError:
            pass
        except httpx.ReadTimeout:
            logger.warning("Prometheus query '{}' timed out", query)
        except httpx.ConnectError:
            logger.error("Prometheus connection failed")
        except httpx.HTTPError as e:
            raise PrometheusException from e

        raise PrometheusException

    async def refresh_data(self) -> None:
        logger.debug("Pulling metrics from Prometheus")
        tasks = []
        for metric in PrometheusMetric:
            tasks.append(asyncio.create_task(self._process_metric(metric)))
        await asyncio.gather(*tasks)

    async def _process_metric(self, metric: PrometheusMetric) -> None:
        data = await self._pull_metric(metric)
        self._data[metric.metric] = metric.transform(data)

    async def close(self) -> None:
        await self._client.aclose()

    def __del__(self) -> None:
        """Clean up client resources on exit."""
        # from aioredis: https://github.com/aio-libs/aioredis-py/blob/ff5a8fe068ebda837d14c3b3777a6182e610854a/aioredis/client.py#L1012
        try:
            loop = asyncio.get_event_loop_policy().get_event_loop()
            if loop.is_running():
                loop.create_task(self.close())
            else:
                loop.run_until_complete(self.close())
        except Exception:
            pass



from fastapi import APIRouter, Depends, WebSocket
from fastapi.exceptions import HTTPException
from starlette.websockets import WebSocketDisconnect

from streams_explorer.api.dependencies.streams_explorer import get_streams_explorer
from streams_explorer.core.services.dataflow_graph import PipelineNotFound
from streams_explorer.models.graph import Graph
from streams_explorer.streams_explorer import StreamsExplorer

router = APIRouter()


@router.get("", response_model=Graph)
async def get_positioned_graph(
    pipeline_name: str | None = None,
    streams_explorer: StreamsExplorer = Depends(get_streams_explorer),
):
    if pipeline_name:
        try:
            return await streams_explorer.get_positioned_pipeline_json_graph(
                pipeline_name
            )
        except PipelineNotFound:
            raise HTTPException(
                status_code=404, detail=f"Pipeline '{pipeline_name}' not found"
            )
    return streams_explorer.get_positioned_json_graph()


@router.websocket("/ws")
async def websocket_endpoint(
    websocket: WebSocket,
    streams_explorer: StreamsExplorer = Depends(get_streams_explorer),
):
    """Send application state updates through WebSocket."""
    await streams_explorer.client_manager.connect(websocket)

    try:
        # bring client up to date
        await streams_explorer.update_client_full(websocket)

        # keep websocket open, allows continuous update
        while websocket:
            await websocket.receive_text()

    except WebSocketDisconnect:
        return
    finally:
        await streams_explorer.client_manager.disconnect(websocket)



from abc import abstractmethod
from collections import defaultdict

from streams_explorer.core.k8s_app import K8sApp
from streams_explorer.models.node_information import NodeInfoListItem
from streams_explorer.plugins import Plugin


class LinkingService(Plugin):
    def __init__(self) -> None:
        self.sink_source_redirects: set[str] = set()
        self._connector_info: list[NodeInfoListItem] = []
        self._streaming_app_info: list[NodeInfoListItem] = []
        self._topic_info: list[NodeInfoListItem] = []
        self._sink_source_info: defaultdict[str, list[NodeInfoListItem]] = defaultdict(
            list
        )

    @property
    def connector_info(self) -> list[NodeInfoListItem]:
        return self._connector_info.copy()

    @connector_info.setter
    def connector_info(self, connector_info: list[NodeInfoListItem]) -> None:
        self._connector_info = connector_info

    @property
    def streaming_app_info(self) -> list[NodeInfoListItem]:
        return self._streaming_app_info.copy()

    @streaming_app_info.setter
    def streaming_app_info(self, streaming_app_info: list[NodeInfoListItem]) -> None:
        self._streaming_app_info = streaming_app_info

    @property
    def topic_info(self) -> list[NodeInfoListItem]:
        return self._topic_info.copy()

    @topic_info.setter
    def topic_info(self, topic_info: list[NodeInfoListItem]) -> None:
        self._topic_info = topic_info

    @property
    def sink_source_info(self) -> dict[str, list[NodeInfoListItem]]:
        return self._sink_source_info.copy()

    @sink_source_info.setter
    def sink_source_info(
        self, sink_source_info: dict[str, list[NodeInfoListItem]]
    ) -> None:
        self._sink_source_info = defaultdict(list, sink_source_info)

    @abstractmethod
    def get_redirect_connector(self, config: dict, link_type: str) -> str | None:
        ...

    @abstractmethod
    def get_redirect_topic(self, topic_name: str, link_type: str) -> str | None:
        ...

    @abstractmethod
    def get_redirect_streaming_app(self, k8s_app: K8sApp, link_type: str) -> str | None:
        ...

    @abstractmethod
    def get_sink_source_redirects(
        self, node_type: str, sink_source_name: str
    ) -> str | None:
        ...

    def add_streaming_app_info_item(self, info_item: NodeInfoListItem) -> None:
        self._streaming_app_info.append(info_item)

    def add_connector_info_item(self, info_item: NodeInfoListItem) -> None:
        self._connector_info.append(info_item)

    def add_topic_info_item(self, info_item: NodeInfoListItem) -> None:
        self._topic_info.append(info_item)

    def add_sink_source_info_item(
        self, sink_source_type: str, info_item: NodeInfoListItem
    ) -> None:
        self._sink_source_info[sink_source_type].append(info_item)



from typing import Awaitable, Callable

from fastapi import FastAPI

from streams_explorer.extractors import load_extractors
from streams_explorer.linker import load_linker
from streams_explorer.metric_provider import load_metric_provider
from streams_explorer.streams_explorer import StreamsExplorer


def setup_default(app: FastAPI) -> Callable[[], Awaitable[None]]:
    async def setup() -> None:
        load_extractors()
        linking_service = load_linker()
        metric_provider = load_metric_provider()
        app.state.streams_explorer = StreamsExplorer(
            linking_service=linking_service(), metric_provider=metric_provider
        )
        await app.state.streams_explorer.setup()

    return setup



from streams_explorer.core.config import settings
from streams_explorer.core.extractor.default.elasticsearch_sink import ElasticsearchSink
from streams_explorer.core.extractor.default.jdbc_sink import JdbcSink
from streams_explorer.core.extractor.default.s3_sink import S3Sink
from streams_explorer.core.extractor.default.streams_bootstrap_producer import (
    StreamsBootstrapProducer,
)
from streams_explorer.core.extractor.extractor import Extractor
from streams_explorer.core.extractor.extractor_container import ExtractorContainer
from streams_explorer.plugins import load_plugin

extractor_container = ExtractorContainer()


def load_default() -> None:
    extractor_container.add(StreamsBootstrapProducer())
    extractor_container.add(ElasticsearchSink())
    extractor_container.add(S3Sink())
    extractor_container.add(JdbcSink())


if settings.plugins.extractors.default:
    load_default()


def load_extractors() -> None:
    extractors = load_plugin(Extractor, all=True)
    for extractor in extractors:
        extractor_container.add(extractor())
    extractor_container.add_generic()



from loguru import logger

from streams_explorer.core.k8s_config_parser import (
    K8sConfigParser,
    StreamsBootstrapEnvParser,
)
from streams_explorer.plugins import load_plugin


def load_config_parser() -> type[K8sConfigParser]:
    parser = load_plugin(K8sConfigParser)  # type: ignore[misc]
    if not parser:
        logger.info("Using default K8sConfigParser")
        return StreamsBootstrapEnvParser
    logger.info(f"Using custom K8sConfigParser: {parser.__name__}")
    return parser
